<?php

namespace Drupal\ui_patterns_field_formatters\Plugin\Derivative;

use Drupal\Component\Plugin\Derivative\DeriverBase;
use Drupal\Component\Plugin\PluginBase;
use Drupal\Core\Entity\EntityFieldManagerInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Entity\EntityTypeRepositoryInterface;
use Drupal\Core\Entity\FieldableEntityInterface;
use Drupal\Core\Field\BaseFieldDefinition;
use Drupal\Core\Field\FieldConfigInterface;
use Drupal\Core\Field\FieldDefinitionInterface;
use Drupal\Core\Field\FieldTypePluginManagerInterface;
use Drupal\Core\Field\FormatterPluginManager;
use Drupal\Core\Logger\LoggerChannelTrait;
use Drupal\Core\Plugin\Context\ContextDefinition;
use Drupal\Core\Plugin\Context\EntityContextDefinition;
use Drupal\Core\Plugin\Discovery\ContainerDeriverInterface;
use Drupal\Core\StringTranslation\StringTranslationTrait;
use Drupal\Core\TypedData\DataDefinitionInterface;
use Drupal\Core\TypedData\TypedDataManagerInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Provides entity field block definitions for every field.
 *
 * @internal
 *   Plugin derivers are internal.
 *
 * @SuppressWarnings(PHPMD.CamelCaseParameterName)
 * @SuppressWarnings(PHPMD.CamelCaseVariableName)
 * @SuppressWarnings(PHPMD.LongClassName)
 * @SuppressWarnings(PHPMD.LongVariable)
 */
abstract class FieldPropertiesSourceDeriverBase extends DeriverBase implements ContainerDeriverInterface {

  use StringTranslationTrait;
  use LoggerChannelTrait;

  /**
   * The entity type repository.
   *
   * @var \Drupal\Core\Entity\EntityTypeRepositoryInterface
   */
  protected EntityTypeRepositoryInterface $entityTypeRepository;

  /**
   * The entity field manager.
   *
   * @var \Drupal\Core\Entity\EntityFieldManagerInterface
   */
  protected EntityFieldManagerInterface $entityFieldManager;

  /**
   * The field type manager.
   *
   * @var \Drupal\Core\Field\FieldTypePluginManagerInterface
   */
  protected FieldTypePluginManagerInterface $fieldTypeManager;

  /**
   * The formatter manager.
   *
   * @var \Drupal\Core\Field\FormatterPluginManager
   */
  protected FormatterPluginManager $formatterManager;

  /**
   * An instance of typed data manager.
   *
   * @var \Drupal\Core\TypedData\TypedDataManagerInterface
   */
  protected TypedDataManagerInterface $typedDataManager;

  /**
   * An instance of entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected EntityTypeManagerInterface $entityTypeManager;

  /**
   * Constructs new FieldBlockDeriver.
   *
   * @param \Drupal\Core\Entity\EntityTypeRepositoryInterface $entity_type_repository
   *   The entity type repository.
   * @param \Drupal\Core\Entity\EntityFieldManagerInterface $entity_field_manager
   *   The entity field manager.
   * @param \Drupal\Core\Field\FieldTypePluginManagerInterface $field_type_manager
   *   The field type manager.
   * @param \Drupal\Core\Field\FormatterPluginManager $formatter_manager
   *   The formatter manager.
   * @param \Drupal\Core\TypedData\TypedDataManagerInterface $typed_data_manager
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   */
  public function __construct(EntityTypeRepositoryInterface $entity_type_repository, EntityFieldManagerInterface $entity_field_manager, FieldTypePluginManagerInterface $field_type_manager, FormatterPluginManager $formatter_manager, TypedDataManagerInterface $typed_data_manager, EntityTypeManagerInterface $entity_type_manager) {
    $this->entityTypeRepository = $entity_type_repository;
    $this->entityFieldManager = $entity_field_manager;
    $this->entityTypeManager = $entity_type_manager;
    $this->fieldTypeManager = $field_type_manager;
    $this->formatterManager = $formatter_manager;
    $this->typedDataManager = $typed_data_manager;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, $base_plugin_id) {
    return new static(
      $container->get('entity_type.repository'),
      $container->get('entity_field.manager'),
      $container->get('plugin.manager.field.field_type'),
      $container->get('plugin.manager.field.formatter'),
      $container->get('typed_data_manager'),
      $container->get('entity_type.manager')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function getDerivativeDefinitions($base_plugin_definition) {
    $entity_type_labels = $this->entityTypeRepository->getEntityTypeLabels();
    foreach ($this->entityFieldManager->getFieldMap() as $entity_type_id => $entity_field_map) {
      foreach ($entity_field_map as $field_name => $field_info) {
        // Skip fields without any formatters.
        $options = $this->formatterManager->getOptions($field_info['type']);
        if (empty($options)) {
          continue;
        }

        foreach ($field_info['bundles'] as $bundle) {
          $derivative = $base_plugin_definition;

          $context_definition = EntityContextDefinition::fromEntityTypeId($entity_type_id)
            ->setLabel($entity_type_labels[$entity_type_id]);
          // We explicitly set the bundle, but won't work if the context value only specifies the entity type.
          $context_definition->addConstraint('Bundle', [$bundle]);

          // We explicitly set the field name, so when props are listed for a field, we only includes field properties
          //   for that field.
          $field_name_context = new ContextDefinition('string');
          $field_name_context->setRequired();
          $field_name_context->setDefaultValue($field_name);
          $field_name_context->addConstraint('AllowedValues', [$field_name]);

          // We explicitly set the bundle, to ensure  when props are listed for a field and a bundle,
          //   we do not list properties from the same field name but from another bundle of the same entity type.
          $bundle_context = new ContextDefinition('string');
          $bundle_context->setRequired();
          $bundle_context->setDefaultValue($bundle);
          $bundle_context->addConstraint('AllowedValues', [$bundle]);

          $derivative['context_definitions'] = [
            'entity' => $context_definition,
            'field_name' => $field_name_context,
            'bundle' => $bundle_context,
          ];
          $this->generateDeriverByEntityBundled($derivative, $entity_type_labels, $entity_type_id, $bundle, $field_name);
        }
      }
    }
    return $this->derivatives;
  }

  /**
   * Generate one deriver given one field of entity bundled.
   *
   * @param array $derivative
   *   The initial array of derivative (used if entity reference is found)
   * @param array $entity_type_labels
   *   The array of entity type labels.
   * @param string $entity_type_id
   *   The entity type id.
   * @param string $bundle
   *   The entity bundle id.
   * @param string $field_name
   *   The machine name of field.
   * @param string $entity_type_id_parent
   *   The entity type of entity parent (case of entity reference).
   * @param $field_definition_parent
   * @param string $field_name_parent
   *   The machine name of field parent (case of entity reference).
   *
   * @return bool
   *   Return false if error.
   *
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  private function generateDeriverByEntityBundled(array $derivative, array $entity_type_labels, string $entity_type_id, string $bundle, string $field_name, string $entity_type_id_parent = '', $field_definition_parent = NULL, string $field_name_parent = ''): bool {
    $field_definitions = $this->entityFieldManager->getFieldDefinitions($entity_type_id, $bundle);
    if (empty($field_definitions[$field_name])) {
      $this->getLogger('field')
        ->error('Field %field_name exists but is missing a corresponding field definition and may be misconfigured.', ['%field_name' => "$entity_type_id.$bundle.$field_name"]);
      return FALSE;
    }
    $field_definition = $field_definitions[$field_name];
    $field_storage_definition = $field_definition->getFieldStorageDefinition();
    $field_property_definitions = $field_storage_definition->getPropertyDefinitions();
    if (empty($field_definition_parent)) {
      $field_definition_parent = $field_definition;
    }

    $this->generateMainPropertyDeriver($derivative, $entity_type_labels, $entity_type_id, $bundle, $field_name, $field_definition);
    foreach ($field_property_definitions as $property_id => $property_definition) {
      // Generate unique deriver ID cause we can't merge contextDefinition,
      // because we can't do OR context constraint.
      $derivative_id = '';
      if (!empty($entity_type_id_parent)) {
        $derivative_id = $entity_type_id_parent . PluginBase::DERIVATIVE_SEPARATOR;
      }
      $derivative_id .= $entity_type_id . PluginBase::DERIVATIVE_SEPARATOR . $bundle . PluginBase::DERIVATIVE_SEPARATOR . $field_name . PluginBase::DERIVATIVE_SEPARATOR . $property_id;
      $derivative['id'] = $derivative_id;
      // label
      // description
      // prop_types
      // context_definitions.
      $derivative['label'] = sprintf('%s %s %s: %s',
        $entity_type_labels[$entity_type_id],
        $bundle,
        // $e*/ntity_type_labels[$entity_type_id], // category
      // Admin label
        $field_definition->getLabel(),
        $property_definition->getLabel(),
      );
      $derivative['description'] = sprintf('%s %s: property %s (%s) of field %s (%s)',
      // Category.
        $entity_type_labels[$entity_type_id],
      // Category.
        $bundle,
        $property_definition->getLabel(),
        $property_id,
        $field_definition->getLabel(),
        $field_name
      );
      $derivative['prop_types'] = $this->getPropTypesDeriver($property_definition);
      // Add a dependency on the field if it is configurable.
      if ($field_definition_parent instanceof FieldConfigInterface) {
        $derivative['config_dependencies'][$field_definition_parent->getConfigDependencyKey()][] = $field_definition_parent->getConfigDependencyName();
      }
      // @todo see if we can skip this config.
      $derivative['ui_patterns_field_formatter'] = [
        'entity_type' => $entity_type_id,
        'bundle' => $bundle,
        'field_name' => $field_name,
        'field_name_parent' => $field_name_parent,
        'property' => $property_id,
        'entity_type_parent' => $entity_type_id_parent,
        'main_property' => FALSE,
        'field_definition' => $field_definition,
        'entity_reference' => $property_definition->getDataType() === 'entity_reference',
      ];

      $this->derivatives[$derivative_id] = $derivative;

      if ($property_definition->getDataType() === 'entity_reference') {
        // @todo Tacke properties from entity referenced from entity_reference fields  
        // $this->generateDeriverEntityReference($entity_type_id, $field_name, $derivative, $field_definition);
      }
    }
    return TRUE;
  }

  /**
   * Generate the deriver for the main property (used for field formatter).
   *
   * @param array $derivative
   * @param array $entity_type_labels
   * @param string $entity_type_id
   * @param string $bundle
   * @param string $field_name
   * @param \Drupal\Core\Field\FieldDefinitionInterface $field_definition
   *
   * @return bool
   */
  private function generateMainPropertyDeriver(array $derivative, array $entity_type_labels, string $entity_type_id, string $bundle, string $field_name, FieldDefinitionInterface $field_definition): bool {
    $field_storage_definition = $field_definition->getFieldStorageDefinition();
    $field_property_definitions = $field_storage_definition->getPropertyDefinitions();
    $main_property = $field_storage_definition->getMainPropertyName();
    // @todo log error.
    if (empty($field_property_definitions[$main_property])) {
      return FALSE;
    }
    $property_definition = $field_property_definitions[$main_property];
    $derivative['id'] = $entity_type_id . PluginBase::DERIVATIVE_SEPARATOR . $bundle . PluginBase::DERIVATIVE_SEPARATOR . $field_name;
    $derivative['label'] = sprintf('%s %s %s',
      $entity_type_labels[$entity_type_id],
      $bundle,
      // $e*/ntity_type_labels[$entity_type_id], // category
    // Admin label
      $field_definition->getLabel()
    );
    $derivative['description'] = sprintf('%s %s : render field %s (%s)',
    // Category.
      $entity_type_labels[$entity_type_id],
      $bundle,
      $field_definition->getLabel(),
      $field_name
    );
    // @todo do we need to restrict on slot only?
    $derivative['prop_types'] = $this->getPropTypesDeriver($property_definition);
    // @todo see if we can skip this config.
    $derivative['ui_patterns_field_formatter'] = [
      'entity_type' => $entity_type_id,
      'bundle' => $bundle,
      'field_name' => $field_name,
      'property' => '',
      'main_property' => TRUE,
      'field_definition' => $field_definition,
    ];

    $this->derivatives[$derivative['id']] = $derivative;
    return TRUE;
  }

  /**
   * Return the mappings of prop type or slots against field drupal type.
   *
   * @param \Drupal\Core\TypedData\DataDefinitionInterface $property_definition
   *   The field property definition.
   *
   * @return array
   *   Array of property mapping between Drupal field and SDC.
   */
  abstract protected function getPropTypesDeriver(DataDefinitionInterface $property_definition): array;

  /**
   * Generate the deriver for entity reference.
   *
   * @param string $entity_type_id_parent
   *   The entity type id of parent.
   * @param string $field_name_parent
   *   The field name of entity referencing.
   * @param array $derivative
   *   The derivative definition of the plugin source?
   * @param \Drupal\Core\Field\FieldDefinitionInterface $field_definition
   *   The field definition.
   *
   * @return bool
   *   False if error.
   *
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  private function generateDeriverEntityReference(string $entity_type_id_parent, string $field_name_parent, array $derivative, FieldDefinitionInterface $field_definition): bool {
    $settings = $field_definition->getSettings();

    // Avoid generate when entity child is the same.
    // @todo maybe need to be change: case of term parent/child.
    if ($settings['target_type'] === $entity_type_id_parent) {
      return FALSE;
    }
    // Get the entity type of field entity reference.
    $entity_type = $this->entityTypeManager->getDefinition($settings['target_type']);
    // Fail with an exception for non-fieldable entity types.
    if (!$entity_type->entityClassImplements(FieldableEntityInterface::class)) {
      return FALSE;
    }
    $entity_type_labels = $this->entityTypeRepository->getEntityTypeLabels();

    // Get the bundle of entity.
    $bundles = $settings['handler_settings']['target_bundles'] ?? [$settings['target_type']];
    // Set parent can be render as main property.
    // $this->derivatives[$derivative['id']]['ui_patterns_field_formatter']['main_property']=TRUE;.
    foreach ($bundles as $bundle) {
      $field_definitions_reference = $this->entityFieldManager->getFieldDefinitions($settings['target_type'], $bundle);
      foreach ($field_definitions_reference as $field_name => $field_info_reference) {
        if ($field_info_reference instanceof BaseFieldDefinition && $field_info_reference->getType() === 'entity_reference') {
          /** @var   \Drupal\Core\Field\TypedData\FieldItemDataDefinition $prop_def */
          $prop_def = $field_info_reference->getFieldStorageDefinition()
            ->getItemDefinition();
          if ($prop_def->getSetting('target_type') === $settings['target_type']) {
            continue;
          }
        }

        $this->generateDeriverByEntityBundled($derivative, $entity_type_labels, $settings['target_type'], $bundle, $field_name, $entity_type_id_parent, $field_definition, $field_name_parent);
      }
    }
    return TRUE;
  }

}
