<?php

declare(strict_types = 1);

namespace Drupal\ui_patterns_field_formatters\Plugin\Field\FieldFormatter;

use Drupal\Core\Field\FieldDefinitionInterface;
use Drupal\Core\Field\FormatterBase;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Plugin\Context\Context;
use Drupal\Core\Plugin\Context\ContextDefinition;
use Drupal\Core\Plugin\Context\ContextRepositoryInterface;
use Drupal\ui_patterns\ComponentPluginManager;
use Drupal\ui_patterns\Form\ComponentFormBuilderTrait;
use Drupal\ui_patterns\Form\ComponentSettingsFormBuilderTrait;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Base class for ui_patterns field formatter plugin.
 *
 * @SuppressWarnings(PHPMD.CamelCaseParameterName)
 * @SuppressWarnings(PHPMD.CamelCaseVariableName)
 * @SuppressWarnings(PHPMD.LongClassName)
 * @SuppressWarnings(PHPMD.LongVariable)
 */
abstract class ComponentFormatterBase extends FormatterBase {

  /**
   * An instance of component plugin manager.
   *
   * @var \Drupal\ui_patterns\ComponentPluginManager
   */
  protected ComponentPluginManager $componentPluginManager;

  /**
   * An instance of context repository plugin.
   *
   * @var \Drupal\Core\Plugin\Context\ContextRepositoryInterface
   */
  protected ContextRepositoryInterface $contextRepository;

  use ComponentSettingsFormBuilderTrait, ComponentFormBuilderTrait {
    ComponentSettingsFormBuilderTrait::getComponentConfiguration insteadof ComponentFormBuilderTrait;
  }

  /**
   * {@inheritdoc}
   */
  public function __construct(
    $plugin_id,
    $plugin_definition,
    FieldDefinitionInterface $field_definition,
    array $settings,
    $label,
    $view_mode,
    array $third_party_settings,
    ComponentPluginManager $component_plugin_manager,
    ContextRepositoryInterface $context_repository
  ) {
    parent::__construct($plugin_id, $plugin_definition, $field_definition, $settings, $label, $view_mode, $third_party_settings);

    $this->componentPluginManager = $component_plugin_manager;
    $this->contextRepository = $context_repository;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    return new static(
      $plugin_id,
      $plugin_definition,
      $configuration['field_definition'],
      $configuration['settings'],
      $configuration['label'],
      $configuration['view_mode'],
      $configuration['third_party_settings'],
      $container->get('plugin.manager.sdc'),
      $container->get('context.repository')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function settingsSummary() {
    $summary = [];
    // Get list of component by manager to insert label.
    $components = $this->componentPluginManager->getDefinitions();
    $options = [];
    $options['_empty'] = $this->t('Empty');
    foreach ($components as $component_id => $component) {
      $options[$component_id] = $component['name'];
    }
    $settings = $this->getSetting('ui_patterns');
    if (!empty($settings['component_id'])) {
      $summary[] = $this->t('Component ":component" selected.', [':component' => $options[$settings['component_id']]]);
    }
    else {
      $summary[] = $this->t('No component selected.');
    }
    return $summary;
  }

  /**
   * {@inheritdoc}
   */
  public static function defaultSettings() {
    return parent::defaultSettings() + self::getComponentFormDefault();
  }

  /**
   * {@inheritdoc}
   */
  public function getComponentSettings(): array {
    if (!empty($this->configuration)) {
      return $this->configuration;
    }
    return $this->getSettings();
  }

  /**
   * {@inheritdoc}
   */
  public function settingsForm(array $form, FormStateInterface $form_state) {
    // @see buildSourceConfigurationForm() of plugin source.
    /** @var \Drupal\ui_patterns\Entity\UiPatternsSampleEntityGenerator $generator */
    return $this->componentSettingsForm($form, $form_state, $this->getComponentSourceContexts());
  }

  /**
   * Set the context of field and entity (override the method trait).
   *
   * @return array
   *   Source contexts.
   */
  protected function getComponentSourceContexts(): array {
    $field_definition = $this->fieldDefinition;
    $entity_type_id = $field_definition->getTargetEntityTypeId();

    $contextFieldNameDefinition = ContextDefinition::create('string');
    // $contextFieldNameDefinition->setRequired(TRUE);
    // $contextFieldNameDefinition->setLabel("Field Name");
    $contextFieldName = new Context($contextFieldNameDefinition, $field_definition->getName());

    $context_source['field_name'] = $contextFieldName;

    $bundle = $field_definition->getTargetBundle();

    $contextBundleDefinition = ContextDefinition::create('string');
    $contextBundle = new Context($contextBundleDefinition, $bundle);
    $context_source['bundle'] = $contextBundle;

    $contexts = $this->contextRepository->getAvailableContexts();
    $runtime_contexts = $this->contextRepository->getRuntimeContexts(array_keys($contexts));
    foreach ($runtime_contexts as $context) {
      if ($context->getContextDefinition()
        ->getDataType() === 'entity:' . $entity_type_id) {
        $context_source['entity'] = $context;
      }
    }

    return $context_source;
  }

  /**
   * Alter the configuration of the mapping, like index and langcode.
   *
   * @param array $specific_config
   *   Array of specific configuration to insert.
   *
   * @return array
   *   The configuration array.
   */
  protected function setComponentSettingsAltered(array $specific_config): array {
    $configuration = $this->getComponentSettings();
    if (!array_key_exists('ui_patterns', $configuration)) {
      return $configuration;
    }
    if (!array_key_exists('props', $configuration['ui_patterns']) && !array_key_exists('slots', $configuration['ui_patterns'])) {
      return $configuration;
    }
    // Loop on each ui patterns props.
    $prop_ids = array_keys($configuration['ui_patterns']['props']);
    foreach ($prop_ids as $prop_id) {
      $configuration['ui_patterns']['props'][$prop_id]['plugin']['ui_patterns_field_formatter'] = $specific_config;
    }
    foreach ($configuration['ui_patterns']['slots'] as $slots_id => $slots) {
      if (empty($slots['sources']) || !is_array($slots['sources'])) {
        continue;
      }
      foreach (array_keys($slots['sources']) as $source_id) {
        $configuration['ui_patterns']['slots'][$slots_id]['sources'][$source_id]['plugin']['ui_patterns_field_formatter'] = $specific_config;
      }
    }

    $this->setComponentConfiguration($configuration['ui_patterns']);
    return $configuration['ui_patterns'];
  }

}
