<?php

declare(strict_types=1);

namespace Drupal\ui_patterns_field_formatters\Plugin\UiPatterns\Source;

use Drupal\Component\Utility\NestedArray;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Field\BaseFieldDefinition;
use Drupal\Core\Field\FieldDefinitionInterface;
use Drupal\Core\Field\FieldStorageDefinitionInterface;
use Drupal\Core\Field\FieldTypePluginManagerInterface;
use Drupal\Core\Field\FormatterInterface;
use Drupal\Core\Field\FormatterPluginManager;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Logger\LoggerChannelTrait;
use Drupal\Core\Plugin\Context\ContextRepositoryInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Component field formatter form builder trait.
 *
 * @SuppressWarnings(PHPMD.CamelCaseParameterName)
 * @SuppressWarnings(PHPMD.CamelCaseVariableName)
 * @SuppressWarnings(PHPMD.LongClassName)
 * @SuppressWarnings(PHPMD.LongVariable)
 */
abstract class FieldFormatterFormSourceBase extends FieldPropertiesSourceBase {

  use LoggerChannelTrait;

  /**
   * {@inheritdoc}
   */
  public static function create(
    ContainerInterface $container,
    array $configuration,
    $plugin_id,
    $plugin_definition
  ) {
    return new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $container->get('context.repository'),
      $container->get('entity_type.manager'),
      $container->get('plugin.manager.field.formatter'),
      $container->get('plugin.manager.field.field_type')

    );
  }

  /**
   * {@inheritdoc}
   */
  public function __construct(
    array $configuration,
    $plugin_id,
    $plugin_definition,
    protected ContextRepositoryInterface $contextRepository,
    protected EntityTypeManagerInterface $entityTypeManager,
    protected FormatterPluginManager $formatterPluginManager,
    protected FieldTypePluginManagerInterface $fieldTypePluginManager,
  ) {
    parent::__construct($configuration, $plugin_id, $plugin_definition, $contextRepository, $entityTypeManager);
  }

  /**
   * {@inheritdoc}
   */
  public function settingsForm(array $form, FormStateInterface $form_state): array {
    $form = parent::settingsForm($form, $form_state);
    $this->buildFieldFormatterForm($form, $form_state);
    return $form;
  }

  /**
   * Test if we can add the formatter field formatter.
   *
   * @return bool|null
   */
  private function isApplicableFormatterFieldFormatter($plugin_definition) {
    if (empty($plugin_definition['ui_patterns_field_formatter']) || empty($plugin_definition['ui_patterns_field_formatter']['field_definition'])) {
      $this->getLogger('ui_patterns_field_formatters')
        ->error('Missing ui_patterns_field_formatter data inside plugin definition %plugin_id', [
          '%plugin_id' => $this->getPluginId(),
        ]);
      return FALSE;
    }
    // Only available for slots.
    $source_name = 'props';
    if (!empty($plugin_definition['prop_types'][0]) && $plugin_definition['prop_types'][0] === 'slot') {
      $source_name = 'slots';
    }
    if ($source_name === 'props') {
      return FALSE;
    }
    // We add field formatter only on the main property field.
    if (empty($plugin_definition['ui_patterns_field_formatter']['main_property'])) {
      return FALSE;
    }
    return TRUE;
  }

  /**
   * Callback to build field formatter form.
   *
   * @param array $form
   *   The source plugin settings form.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *
   * @return void|null
   *
   * @throws \Drupal\Component\Plugin\Exception\PluginException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  private function buildFieldFormatterForm(array &$form, FormStateInterface $form_state) {
    $plugin_definition = $this->getPluginDefinition();
    if (!$this->isApplicableFormatterFieldFormatter($plugin_definition)) {
      return NULL;
    }
    /** @var  \Drupal\Core\Field\FieldDefinitionInterface $field_definition */
    $field_definition = $plugin_definition['ui_patterns_field_formatter']['field_definition'];

    if (!$field_definition instanceof FieldDefinitionInterface) {
      return NULL;
    }
    $field_storage = $field_definition->getFieldStorageDefinition();
    $formatter_options = $this->getAvailableFormatterOptions($field_storage, $field_definition);

    if (empty($formatter_options)) {
      return NULL;
    }
    // Get the formatter type from configuration.
    $type_path = [
      'settings',
      'type',
    ];
    $formatter_type = $this->getSettingsFromConfiguration($type_path);

    // Get the formatter settings from configuration.
    if (!empty($formatter_type)) {
      $settings_path = [
        'settings',
        'settings',
      ];
      $settings = $this->getSettingsFromConfiguration($settings_path);
    }
    // Get default formatter type.
    if (empty($formatter_type) || !isset($formatter_options[$formatter_type])) {
      $formatter_type = $this->fieldTypePluginManager->getDefinition($field_storage->getType())['default_formatter'] ?? key($formatter_options);
      $settings = $this->formatterPluginManager->getDefaultSettings($field_storage->getType());
    }
    // Reset settings if we change the formatter.
    $trigerring_element = $form_state->getTriggeringElement();
    if (!empty($trigerring_element) && $trigerring_element['#value'] === $formatter_type) {
      $settings = $this->formatterPluginManager->getDefaultSettings($formatter_type);
    }
    if (empty($settings) && !empty($formatter_type)) {
      $settings = $this->formatterPluginManager->getDefaultSettings($formatter_type);
    }
    $this->generateFieldFormatterForm($form, $form_state, $formatter_options, $settings, $field_storage, $formatter_type);
  }

  /**
   * Generate the form formatter field formatter.
   *
   * @param array $form
   *   The source plugin settings form.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current form state.
   * @param array $formatter_options
   *   The doubting formatters options.
   * @param mixed $settings
   *   The settings of formatter field formatter.
   * @param \Drupal\Core\Field\FieldStorageDefinitionInterface $field_storage
   * @param string $formatter_type
   *   The chosen or default formatter.
   *
   * @return void
   */
  private function generateFieldFormatterForm(array &$form, FormStateInterface $form_state, array $formatter_options, mixed $settings, FieldStorageDefinitionInterface $field_storage, string $formatter_type): void {
    $form['type'] = [
      '#type' => 'select',
      '#title' => $this->t('Formatter'),
      '#options' => $formatter_options,
      '#default_value' => $formatter_type,
      // Note: We cannot use ::foo syntax, because the form is the entity form
      // display.
      '#ajax' => [
        'callback' => [__CLASS__, 'onFormatterTypeChange'],
        'wrapper' => 'field-formatter-settings-ajax',
        'method' => 'replace',
      ],
    ];

    $options = [
      'field_definition' => $this->getFieldDefinition($field_storage),
      'configuration' => [
        'type' => $formatter_type,
        'settings' => $settings,
        'label' => '',
        'weight' => 0,
      ],
      'view_mode' => '_custom',
    ];

    // Get the formatter settings form.
    $settings_form = ['#value' => []];
    if ($formatter = $this->formatterPluginManager->getInstance($options)) {
      $settings_form = $formatter->settingsForm([], $form_state);
    }

    $form['settings'] = $settings_form;
    $form['settings']['#prefix'] = '<div id="field-formatter-settings-ajax">' . ($form['settings']['#prefix'] ?? '');
    $form['settings']['#suffix'] = ($form['settings']['#suffix'] ?? '') . '</div>';
  }

  /**
   * Get the settings from the plugin configuration.
   *
   * @param array $setting
   *
   * @return mixed|null
   */
  protected function getSettingsFromConfiguration(array $setting) {
    $configuration = $this->getConfiguration();
    return NestedArray::getValue($configuration, $setting);
  }

  /**
   * Get field definition for given field storage definition.
   *
   * @param \Drupal\Core\Field\FieldStorageDefinitionInterface $field_storage_definition
   *   The field storage definition.
   *
   * @return \Drupal\Core\Field\BaseFieldDefinition
   *   The field definition.
   */
  protected function getFieldDefinition(FieldStorageDefinitionInterface $field_storage_definition): BaseFieldDefinition {
    return BaseFieldDefinition::createFromFieldStorageDefinition($field_storage_definition);
  }

  /**
   * Create an instance of field formatter.
   *
   * @param string $formatter_id
   *   The formatter id.
   * @param \Drupal\Core\Field\FieldDefinitionInterface $field_definition
   *   The field definition of field to apply formatter.
   *
   * @return \Drupal\Core\Field\FormatterBase
   *   The field formatter plugin.
   *
   * @throws \Drupal\Component\Plugin\Exception\PluginException
   */
  private function createInstanceFormatter(string $formatter_id, FieldDefinitionInterface $field_definition) {
    // @todo Ensure it is right to empty all values here, see:
    // https://api.drupal.org/api/drupal/core%21lib%21Drupal%21Core%21Field%21FormatterPluginManager.php/class/FormatterPluginManager/8.2.x
    $configuration = [
      'field_definition' => $field_definition,
      'settings' => [],
      'label' => '',
      'view_mode' => '',
      'third_party_settings' => [],
    ];
    return $this->formatterPluginManager->createInstance($formatter_id, $configuration);
  }

  /**
   * Filter the field formatter plugin given the field definition.
   *
   * @param \Drupal\Core\Field\FormatterBase[] $formatter_instances
   *   Array of field formatter plugin.
   * @param \Drupal\Core\Field\FieldDefinitionInterface $field_definition
   *
   * @return array
   */
  private function filterFormatter(array $formatter_instances, FieldDefinitionInterface $field_definition): array {
    $filtered = [];
    foreach ($formatter_instances as $formatter) {
      if (!$formatter instanceof FormatterInterface || !$formatter::isApplicable($field_definition)) {
        continue;
      }
      $filtered[$formatter->getPluginId()] = $formatter;
    }
    return $filtered;
  }

  /**
   * Get all available formatters by loading available ones and filtering out.
   *
   * @param \Drupal\Core\Field\FieldStorageDefinitionInterface $field_storage_definition
   *   The field storage definition.
   * @param \Drupal\Core\Field\FieldDefinitionInterface $field_definition
   *
   * @return string[]
   *   The field formatter labels keys by plugin ID.
   *
   * @throws \Drupal\Component\Plugin\Exception\PluginException
   */
  protected function getAvailableFormatterOptions(FieldStorageDefinitionInterface $field_storage_definition, FieldDefinitionInterface $field_definition): array {
    $formatters = $this->formatterPluginManager->getOptions($field_storage_definition->getType());
    $formatter_instances = [];
    foreach ($formatters as $formater_id => $formatter) {
      $formatter_instances[$formater_id] = $this->createInstanceFormatter($formater_id, $field_definition);
    }

    $filtered_formatter_instances = $this->filterFormatter($formatter_instances, $field_definition);
    $options = array_map(
      static function (FormatterInterface $formatter) {
        return $formatter->getPluginDefinition()['label'];
      }, $filtered_formatter_instances);

    // Remove field_link itself.
    unset($options['field_link']);

    return $options;
  }

  /**
   * Ajax callback for fields with AJAX callback to update form substructure.
   *
   * @param array $form
   *   The form.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The form state.
   *
   * @return array
   *   The replaced form substructure.
   */
  public static function onFormatterTypeChange(array $form, FormStateInterface $form_state) {
    $triggeringElement = $form_state->getTriggeringElement();
    // Dynamically return the dependent ajax for elements based on the
    // triggering element. This shouldn't be done statically because
    // settings forms may be different, e.g. for layout builder, core, ...
    if (!empty($triggeringElement['#array_parents'])) {
      $subformKeys = $triggeringElement['#array_parents'];
      // Remove the triggering element itself and add the 'settings' below key.
      array_pop($subformKeys);
      $subformKeys[] = 'settings';
      // Return the subform:
      return NestedArray::getValue($form, $subformKeys);
    }
    return [];
  }

}
