<?php

declare(strict_types=1);

namespace Drupal\ui_patterns_field_formatters\Plugin\UiPatterns\Source;

use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Field\FieldItemListInterface;
use Drupal\Core\Field\Plugin\Field\FieldType\EntityReferenceItemInterface;
use Drupal\Core\Logger\LoggerChannelTrait;
use Drupal\Core\TypedData\TypedDataInterface;
use Drupal\ui_patterns\SourceInterface;

/**
 * Class helper to build field formatter of value.
 *
 * @SuppressWarnings(PHPMD.CamelCaseParameterName)
 * @SuppressWarnings(PHPMD.CamelCaseVariableName)
 * @SuppressWarnings(PHPMD.LongClassName)
 * @SuppressWarnings(PHPMD.LongVariable)
 */
abstract class FieldPropertiesSourceBase extends UiPatternsFieldFormattersSourceBase implements SourceInterface {

  use LoggerChannelTrait;

  /**
   * Extract the field of entity.
   *
   * @return \Drupal\Core\Field\FieldItemListInterface|mixed|null
   *   Return the field items of entity.
   *
   * @throws \Drupal\Component\Plugin\Exception\ContextException
   */
  protected function getFieldEntity() {
    // Get data from the plugin definition.
    // @see inside FieldPropertiesSourceDeriverBase the method
    // PgenerateDeriverByEntityBundled().
    $field_name = $this->getPluginDefValue('field_name');
    if (empty($field_name)) {
      return NULL;
    }
    $field_name_parent = $this->getPluginDefValue('field_name_parent', FALSE);

    /** @var  \Drupal\Core\Entity\ContentEntityBase $entity */
    $entity = $this->getContextValue('entity');
    // Case of entity reference.
    if (!empty($field_name_parent)) {
      if (!$entity->hasField($field_name_parent)) {
        $this->getLogger('ui_patterns_field_formatters')
          ->error('Entity %entity_type %bundle has no field %field_name', [
            '%entity_type' => $entity->getEntityTypeId(),
            '%bundle' => $entity->bundle(),
            '%field_name' => $field_name_parent,
          ]);
        return NULL;
      }
      // Return the entity child.
      return $entity->get($field_name_parent);
    }
    if (!$entity->hasField($field_name)) {
      $this->getLogger('ui_patterns_field_formatters')
        ->error('Entity %entity_type %bundle has no field %field_name', [
          '%entity_type' => $entity->getEntityTypeId(),
          '%bundle' => $entity->bundle(),
          '%field_name' => $field_name,
        ]);
      return NULL;
    }
    /** @var  \Drupal\Core\Field\FieldItemList $field */
    return $entity->get($field_name);
  }

  /**
   * Extract the field value.
   *
   * @param int $field_index
   *   The delta of field to get the value.
   * @param \Drupal\Core\Field\FieldItemListInterface|null $items
   *   The items of the fields.
   *
   * @return mixed|null
   *   The field value.
   *
   * @throws \Drupal\Component\Plugin\Exception\ContextException
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   * @throws \Drupal\Core\TypedData\Exception\MissingDataException
   */
  protected function getFieldIndexedValueEntity(int $field_index, FieldItemListInterface|null $items = NULL): mixed {
    $field_value = $this->getFieldValue($field_index, $items);
    if (!$field_value instanceof EntityReferenceItemInterface) {
      return $field_value->getValue();
    }
    $target_id = $this->getEntityReferenceId($field_value);
    // Case of entity reference.
    $entity_type = $this->getPluginDefValue('entity_type');
    $field_name = $this->getPluginDefValue('field_name');
    if (empty($entity_type) || empty($field_name) || $target_id === FALSE) {
      return NULL;
    }
    $items = $this->loadFieldEntityReference($entity_type, $target_id, $field_name);
    return $items->get($field_index)->getValue();
  }

  /**
   * Return the field item typed data.
   *
   * @param int $field_index
   *   The delta of field to get the value.
   * @param \Drupal\Core\Field\FieldItemListInterface|null $items
   *   The items of the fields.
   *
   * @return \Drupal\Core\TypedData\TypedDataInterface|bool|null
   *   The field item typed data.
   *
   * @throws \Drupal\Component\Plugin\Exception\ContextException
   * @throws \Drupal\Core\TypedData\Exception\MissingDataException
   */
  protected function getFieldValue(int $field_index, FieldItemListInterface|null $items = NULL): TypedDataInterface|null|bool {
    // Get the field of entity.
    if ($items === NULL) {
      $items = $this->getFieldEntity();
    }
    if (empty($items)) {
      return FALSE;
    }
    $field_value = $items->get($field_index);
    if (!$field_value) {
      $this->getLogger('ui_patterns_field_formatters')
        ->error('Index %index not found on field %field_name', [
          '%index' => $field_index,
          '%field_name' => $items->getName(),
        ]);
      return FALSE;
    }
    return $field_value;
  }

  /**
   * Get the entity reference id from the typed data.
   *
   * @param \Drupal\Core\TypedData\TypedDataInterface $field_value
   *
   * @return string|bool
   *   The entity reference id or FALSE if it's not entity reference.
   */
  protected function getEntityReferenceId(TypedDataInterface $field_value): string|bool {
    if ($field_value instanceof EntityReferenceItemInterface) {
      $direct_value = $field_value->getValue();
      return $direct_value['target_id'];
    }
    return FALSE;
  }

  /**
   * Load the field of the entity reference.
   *
   * @param string $entity_type
   *   The entity type of entity reference to load.
   * @param $entity_id
   *   The id of entity to load.
   * @param string $field_name
   *   The field name of entity to get data.
   *
   * @return mixed|null
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  protected function loadFieldEntityReference(string $entity_type, $entity_id, string $field_name): mixed {
    // Load the entity.
    $entity = $this->entityTypeManager->getStorage($entity_type)
      ->load($entity_id);
    if (!$entity instanceof EntityInterface) {
      return NULL;
    }
    if (!$entity->hasField($field_name)) {
      $this->getLogger('ui_patterns_field_formatters')
        ->error('Entity %entity_type %bundle has no field %field_name', [
          '%entity_type' => $entity->getEntityTypeId(),
          '%bundle' => $entity->bundle(),
          '%field_name' => $field_name,
        ]);
      return NULL;
    }
    // Case of all formatter.
    return $entity->get($field_name);
  }

}
