<?php

declare(strict_types=1);

namespace Drupal\ui_patterns_field_formatters\Plugin\UiPatterns\Source;

use Drupal\Core\Field\FieldItemBase;
use Drupal\Core\Field\FieldItemListInterface;
use Drupal\Core\Field\Plugin\Field\FieldType\EntityReferenceItemInterface;

/**
 * Plugin implementation of the field_label source.
 *
 * @Source(
 *   id = "field_properties_source_prop",
 *   label = @Translation("Field Property (Props)"),
 *   description = @Translation("Field property source plugin for props."),
 *   deriver =
 *   "Drupal\ui_patterns_field_formatters\Plugin\Derivative\FieldPropertiesSourcePropDeriver",
 * )
 *
 * @SuppressWarnings(PHPMD.CamelCaseParameterName)
 * @SuppressWarnings(PHPMD.CamelCaseVariableName)
 * @SuppressWarnings(PHPMD.LongClassName)
 * @SuppressWarnings(PHPMD.LongVariable)
 */
class FieldPropertiesSourceProp extends FieldPropertiesSourceBase {

  /**
   * {@inheritdoc}
   */
  public function getDataInternal(array $ui_patterns_field_formatters_config, FieldItemListInterface|null $items = NULL): mixed {
    if ($items === NULL) {
      $items = $this->getFieldEntity();
    }

    $plugin_definition = $this->getPluginDefinition();
    if (empty($plugin_definition['ui_patterns_field_formatter'])) {
      $this->getLogger('ui_patterns_field_formatters')
        ->error('Missing ui_patterns_field_formatter data inside plugin definition %plugin_id', [
          '%plugin_id' => $this->getPluginId(),
        ]);
      return NULL;
    }
    // Get field index or 0 (first if formatter all).
    $index_item = !empty($ui_patterns_field_formatters_config['field_index']) ? $ui_patterns_field_formatters_config['field_index'] : 0;
    $property = $this->getPluginDefValue('property');
    if (empty($property)) {
      return $this->getDataMainproperty($items, $index_item);
    }
    $field_value = $this->getFieldIndexedValueEntity($index_item, $items);
    if (!empty($field_value) && array_key_exists($property, $field_value)) {
      return $this->transTypeProp($field_value[$property], $this->propId);
    }
    return NULL;
  }

  /**
   * Get the data from the main property.
   *
   * @param \Drupal\Core\Field\FieldItemListInterface $items
   *   The items of the field.
   * @param int $index_item
   *   The delta of field to get the value.
   *
   * @return bool|float|int|mixed|null
   *   The data from the main property.
   *
   * @throws \Drupal\Component\Plugin\Exception\ContextException
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   * @throws \Drupal\Core\TypedData\Exception\MissingDataException
   */
  protected function getDataMainproperty(FieldItemListInterface $items, int $index_item) {
    $field_value = $this->getFieldValue($index_item, $items);
    if ($field_value instanceof EntityReferenceItemInterface) {
      $target_id = $this->getEntityReferenceId($field_value);
      // Case of entity reference.
      $entity_type = $this->getPluginDefValue('entity_type');
      $field_name = $this->getPluginDefValue('field_name');
      if (empty($entity_type) || empty($field_name) || $target_id === FALSE) {
        return NULL;
      }
      $items = $this->loadFieldEntityReference($entity_type, $target_id, $field_name);
    }
    /** @var  \Drupal\Core\Field\FieldItemBase $item */
    $item = $items->get($index_item);
    if (!$item instanceof FieldItemBase) {
      return NULL;
    }
    $configuration = $this->getConfiguration();

    // Field formatter is set.
    if (!empty($configuration['settings']['type'])) {
      // @todo pb trimmed balise html?
      return $this->transTypeProp($item->view([
        'type' => $configuration['settings']['type'],
        'settings' => $configuration['settings']['settings'] ?? [],
      ]), $this->propId);
    }
    return $this->transTypeProp($item->view(), $this->propId);
  }

  /**
   * Trans-type the value to be valid inside SDC.
   *
   * @param mixed $value
   *   The value to trans-type.
   * @param string $prop_id
   *   The prop drupal id.
   *
   * @return bool|float|int
   *   The value converted.
   *
   * @todo this part will be move into UI patterns.
   */
  private function transTypeProp(mixed $value, string $prop_id): mixed {
    return match ($prop_id) {
      'decimal', 'integer', 'duration' => (int) $value,
      'float' => (float) $value,
      'boolean' => (boolean) $value,
      default => $value,
    };
  }

}
