<?php

declare(strict_types=1);

namespace Drupal\ui_patterns_field_formatters\Plugin\UiPatterns\Source;

use Drupal\Core\Entity\Plugin\DataType\EntityReference;
use Drupal\Core\Field\FieldItemInterface;
use Drupal\Core\Field\FieldItemListInterface;
use Drupal\Core\TypedData\Plugin\DataType\Uri;
use Drupal\Core\Url;
use Drupal\text\TextProcessed;

/**
 * Plugin implementation of the field_label source.
 *
 * @Source(
 *   id = "field_properties_source_slot",
 *   label = @Translation("Field Property (Slots)"),
 *   description = @Translation("Field property source plugin for slot."),
 *   deriver =
 *   "Drupal\ui_patterns_field_formatters\Plugin\Derivative\FieldPropertiesSourceSlotDeriver",
 * )
 *
 * @SuppressWarnings(PHPMD.CamelCaseParameterName)
 * @SuppressWarnings(PHPMD.CamelCaseVariableName)
 * @SuppressWarnings(PHPMD.LongClassName)
 * @SuppressWarnings(PHPMD.LongVariable)
 */
class FieldPropertiesSourceSlot extends FieldFormatterFormSourceBase {

  /**
   * {@inheritdoc}
   */
  public function getDataInternal(array $ui_patterns_field_formatters_config, FieldItemListInterface|null $items = NULL): mixed {
    if ($items === NULL) {
      $items = $this->getFieldEntity();
    }
    $final_value = [];
    if (!$items instanceof FieldItemListInterface) {
      return [];
    }
    // Render main property.
    if ($this->isApplicableFieldFormatter()) {
      $delta_field = $ui_patterns_field_formatters_config['field_index'] ?? NULL;
      $this->extractValueFieldFormatter($items, $final_value, $delta_field);
      return $final_value;
    }
    $property = $this->getPluginDefValue('property');
    if (empty($property)) {
      return [];
    }
    $lang_code = $ui_patterns_field_formatters_config['lang_code'] ?? 'und';
    return $this->getDataItems($items, $property, $lang_code, $ui_patterns_field_formatters_config['field_index'] ?? NULL);
  }

  /**
   * @param \Drupal\Core\Field\FieldItemListInterface $items
   * @param $property
   * @param $lang_code
   *
   * @return mixed
   * @throws \Drupal\Component\Plugin\Exception\ContextException
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   * @throws \Drupal\Core\TypedData\Exception\MissingDataException
   */
  private function getDataItems(FieldItemListInterface $items, $property, $lang_code, int|null $field_index): mixed {
    $final_value = [];
    $item_definition = $items->getFieldDefinition();
    $field_type = $item_definition->get('field_type');
    $field_name_parent = $this->getPluginDefValue('field_name_parent', FALSE);
    if ($field_type !== 'entity_reference' || empty($field_name_parent)) {
      // We take only the first value -> TODO?
      $this->generateFinaleValue($items, $property, $field_index, $final_value, $lang_code);
      return $final_value;
    }
    $entity_type = $this->getPluginDefValue('entity_type');
    if (empty($entity_type)) {
      return [];
    }
    $field_name = $this->getPluginDefValue('field_name');
    if (empty($field_name)) {
      return [];
    }
    foreach ($items as $delta => $item) {
      if ($field_index !== NULL && $delta != $field_index) {
        continue;
      }
      $field_target_value = $item->getValue();
      // Load the field of entity reference.
      $items_ref = $this->loadFieldEntityReference($entity_type, $field_target_value['target_id'], $field_name);
      if ($items_ref instanceof FieldItemListInterface) {
        $this->generateFinaleValue($items_ref, $property, $field_index, $final_value, $lang_code);
      }
    }
    return $final_value;
  }

  /**
   * Generate the render array of all formatter.
   *
   * @param \Drupal\Core\Field\FieldItemListInterface $items
   *   The field items.
   * @param string $property
   *   The property field key.
   * @param array $final_value
   *   The render array of value to display.
   * @param string $lang_code
   *   The lang code.
   *
   * @return void
   *
   * @throws \Drupal\Core\TypedData\Exception\MissingDataException
   */
  private function generateFinaleValue(FieldItemListInterface $items, string $property, int|null $field_index, array &$final_value = [], string $lang_code = ''): void {
    foreach ($items->getIterator() as $delta => $fieldValue) {
      if ($field_index !== NULL && $delta != $field_index) {
        continue;
      }
      $final_value[] = ['#markup' => $this->extractValue($fieldValue, $property, $lang_code)];
    }
  }

  /**
   * Extract the given value from the item.
   *
   * @param \Drupal\Core\Field\FieldItemInterface $item
   *   The field item.
   * @param string $source
   *   The source property.
   * @param string $langcode
   *   The language that should be used to render the field.
   *
   * @return mixed|null
   *   The extracted value or NULL if the field is referencing an entity that
   *   doesn't exist anymore.
   *
   * @throws \Drupal\Core\TypedData\Exception\MissingDataException
   */
  protected function extractValue(FieldItemInterface $item, string $source, string $langcode): mixed {
    $property = $item->get($source);
    if ($property instanceof TextProcessed) {
      return $property->getValue();
    }
    if ($property instanceof EntityReference) {
      // Ensure the referenced entity still exists.
      if (empty($property->getTarget())) {
        return NULL;
      }

      $target = $property->getTarget();
      if (empty($target)) {
        return NULL;
      }
      $entity = $target->getEntity();
      // Drupal loads the entity in its default language and should load
      // the translated one if available.
      $value = $entity->label();
      if ($entity->hasTranslation($langcode)) {
        $translated_entity = $entity->getTranslation($langcode);
        $value = $translated_entity->label();
      }
      return $value;
    }
    $value = $property->getValue();
    $value = empty($value) ? '' : $value;

    // Preprocess Uri datatype instead of pushing the raw value.
    if ($property instanceof Uri) {
      $options = [];
      // Most of the time, Uri datatype is met in a "link" field type.
      if ($item->getPluginId() === 'field_item:link') {
        $options = $item->get('options')->toArray();
      }
      return Url::fromUri($value, $options)->toString();
    }
    return $value;
  }

  /**
   * Test if the field formatter settings has been set.
   *
   * @return bool
   *   Return true if we need to apply field formatter.
   */
  private function isApplicableFieldFormatter(): bool {
    $plugin_definition = $this->getPluginDefinition();
    if (empty($plugin_definition['ui_patterns_field_formatter'])) {
      $this->getLogger('ui_patterns_field_formatters')
        ->error('Missing ui_patterns_field_formatter data inside plugin definition %plugin_id', [
          '%plugin_id' => $this->getPluginId(),
        ]);
      return FALSE;
    }

    // Not the main property.
    if (empty($plugin_definition['ui_patterns_field_formatter']['main_property'])) {
      return FALSE;
    }

    $configuration = $this->getConfiguration();
    // No field formatter set.
    if (empty($configuration['settings']['type'])) {
      return FALSE;
    }
    return TRUE;
  }

  /**
   * Render the field with the field formatter.
   *
   * @param \Drupal\Core\Field\FieldItemListInterface $items
   * @param array $final_value
   * @param int|null $field_delta
   *
   * @return void
   */
  private function extractValueFieldFormatter(FieldItemListInterface $items, array &$final_value, int|null $field_delta = NULL): void {
    $configuration = $this->getConfiguration();
    if (empty($configuration['settings']['type'])) {
      return;
    }
    foreach ($items->getIterator() as $delta => $item) {
      if ($field_delta !== NULL) {
        if ($delta !== $field_delta) {
          continue;
        }

        $final_value[] = $item->view([
          'type' => $configuration['settings']['type'],
          'settings' => $configuration['settings']['settings'] ?? [],
        ]);
        break;
      }
      $final_value[] = $item->view([
        'type' => $configuration['settings']['type'],
        'settings' => $configuration['settings']['settings'],
      ]);
    }
  }

}
