<?php

declare(strict_types=1);

namespace Drupal\ui_patterns_field_formatters\Plugin\UiPatterns\Source;

use Drupal\Component\Utility\NestedArray;
use Drupal\Component\Utility\Xss;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Field\FieldItemListInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Logger\LoggerChannelTrait;
use Drupal\Core\Plugin\Context\ContextRepositoryInterface;
use Drupal\Core\StringTranslation\TranslatableMarkup;
use Drupal\ui_patterns\SourceInterface;
use Drupal\ui_patterns\SourcePluginBase;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Twig\Environment;

/**
 * Base class for ui_patterns_field_formatter source plugin.
 *
 *   @SuppressWarnings(PHPMD.CamelCaseParameterName)
 *   @SuppressWarnings(PHPMD.CamelCaseVariableName)
 *   @SuppressWarnings(PHPMD.LongClassName)
 *   @SuppressWarnings(PHPMD.LongVariable)
 */
abstract class UiPatternsFieldFormattersSourceBase extends SourcePluginBase implements SourceInterface {

  use LoggerChannelTrait;

  /**
   * {@inheritdoc}
   */
  public static function create(
    ContainerInterface $container,
    array $configuration,
    $plugin_id,
    $plugin_definition
  ) {
    return new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $container->get('context.repository'),
      $container->get('entity_type.manager')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function __construct(
    array $configuration,
    $plugin_id,
    $plugin_definition,
    protected ContextRepositoryInterface $contextRepository,
    protected EntityTypeManagerInterface $entityTypeManager
  ) {
    parent::__construct($configuration, $plugin_id, $plugin_definition, $contextRepository);
  }

  /**
   * Get data inside plugin definition.
   *
   * @param string $key
   *   The key name of plugin definition to get data.
   * @param bool $log_error
   *   If true, log error when the key doesn't exist inside plugin definition.
   *
   * @return false|mixed
   *   The data inside plugin definition or false if error.
   */
  protected function getPluginDefValue(string $key, bool $log_error = TRUE): mixed {
    $plugin_definition = $this->getPluginDefinition();
    if (empty($plugin_definition)) {
      $this->getLogger('ui_patterns_field_formatters')
        ->error('The plugin definition of %plugin_id  is empty', [
          '%plugin_id' => $this->getPluginId(),
        ]);
      return FALSE;
    }
    if (empty($plugin_definition['ui_patterns_field_formatter']) || empty($plugin_definition['ui_patterns_field_formatter'][$key])) {
      if ($log_error) {
        $this->getLogger('ui_patterns_field_formatters')
          ->error('Missing ui_patterns_field_formatter data (%key) inside plugin definition %plugin_id', [
            '%key' => $key,
            '%plugin_id' => $this->getPluginId(),
          ]);
      }
      return FALSE;
    }
    return $plugin_definition['ui_patterns_field_formatter'][$key];
  }

  /**
   * {@inheritdoc}
   */
  public function settingsForm(array $form, FormStateInterface $form_state): array {
    $form = parent::settingsForm($form, $form_state);
    $form['label_map'] = [
      "#type" => "label",
      "#title" => $this->propDefinition['title'] . ": " . $this->label(),
    ];
    $field_name = $this->getContextValue('field_name');
    $configuration = $this->getConfiguration();

    $prop_id = $configuration['prop_id'] ?? NULL;
    $plugin_definition = $this->getPluginDefinition();
    if ($plugin_definition['description'] instanceof TranslatableMarkup) {
      $plugin_definition['description'] = $plugin_definition['description']->render();
    }
    $source_name = 'props';
    if (!empty($plugin_definition['prop_types'][0]) && $plugin_definition['prop_types'][0] === 'slot') {
      $source_name = 'slots';
    }
    // Add feature to alter field.
    $this->buildAlterFieldForm($form, $plugin_definition['description'] ?? '', $source_name, $prop_id, $field_name);
    return $form;
  }

  /**
   * Build the settings form for alter field feature.
   *
   * @param array $form
   *   The render array of settings form.
   * @param string $plugin_description
   *   The plugin description.
   * @param string $source_name
   *   The source type (slots or props).
   * @param string $prop_id
   * @param string $field_name
   *
   * @return array|void
   */
  private function buildAlterFieldForm(array &$form, string $plugin_description, string $source_name, string $prop_id, string $field_name) {
    // We add alter option only for props type string.
    $prop_id_allowed = ['string', 'string_default'];
    if ($source_name === 'props' && !in_array($prop_id, $prop_id_allowed, TRUE)) {
      return $form;
    }
    // Add alter field form option.
    $form['alter']['#tree'] = TRUE;
    $alter_path = ['settings', 'alter'];
    $alter_default_value = $this->getSettingsFromConfiguration($alter_path);
    $alter_default_value = !empty($alter_default_value) ? $alter_default_value : [];
    $form['alter']['alter_field'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Override the output of this field with custom text'),
      '#default_value' => array_key_exists('alter_field', $alter_default_value) ? $alter_default_value['alter_field'] : FALSE,
    ];
    // Build the states Form API key.
    // Key for props.
    $checkbox_states = '"fields[' . $field_name . '][settings_edit_form][settings][ui_patterns][component_form][' . $source_name . '][' . $prop_id . '][plugin][alter][alter_field]"';
    // Key for slots.
    if ($source_name === 'slots') {
      $checkbox_states = '"fields[' . $field_name . '][settings_edit_form][settings][ui_patterns][component_form][' . $source_name . '][' . $prop_id . '][sources][0][plugin][alter][alter_field]"';
    }
    $form['alter']['text'] = [
      '#title' => $this->t('Text'),
      '#type' => 'textarea',
      '#default_value' => array_key_exists('text', $alter_default_value) ? $alter_default_value['text'] : '',
      // The tag list will be escaped.
      '#description' => $this->t('The text to display for this field. You may enter data from this @source_name as per the "Replacement patterns" below. You may include <a href="@twig_docs">Twig</a> or the following allowed HTML tags: <code>@tags</code>', [
        '@source_name' => $source_name,
        '@twig_docs' => 'https://twig.symfony.com/doc/' . Environment::MAJOR_VERSION . '.x',
        '@tags' => '<' . implode('> <', Xss::getAdminTagList()) . '>',
      ]),
      '#states' => [
        'visible' => [
          ':input[name=' . $checkbox_states . ']' => ['checked' => TRUE],
        ],
      ],
    ];
    // Build data replacements.
    $output = [];
    if (!empty($prop_id)) {
      $output[] = [
        '#markup' => '<p>' . $this->t("The following replacement tokens are available for this field.") . '</p>',
      ];
      $items = [];
      $items[] = '{{' . $prop_id . '}}' . ' :' . $plugin_description . ' render inside ' . $prop_id;
      $item_list = [
        '#theme' => 'item_list',
        '#items' => $items,
      ];
      $output[] = $item_list;
    }

    $form['alter']['help'] = [
      '#type' => 'details',
      '#title' => $this->t('Replacement patterns'),
      '#value' => $output,
      '#states' => [
        'visible' => [
          ':input[name=' . $checkbox_states . ']' => ['checked' => TRUE],
        ],
      ],
    ];
  }

  /**
   * Get the data of slot or prop.
   *
   * @param array $ui_patterns_field_formatters_config
   * @param \Drupal\Core\Field\FieldItemListInterface|null $items
   *
   * @return mixed
   */
  abstract public function getDataInternal(array $ui_patterns_field_formatters_config, FieldItemListInterface|null $items = NULL): mixed;

  /**
   * {@inheritdoc}
   */
  public function getPropValue(): mixed {
    $config = $this->configuration;
    $config_form_value = array_key_exists('settings', $config) ? $config['settings'] : [];
    $ui_patterns_field_formatters_config = array_key_exists('ui_patterns_field_formatter', $config_form_value) ? $config_form_value['ui_patterns_field_formatter'] : [];
    $items = $this->getFieldItems();
    $data = $this->getDataInternal($ui_patterns_field_formatters_config, $items);
    return $this->alterField($this->getConfiguration(), $data);
  }

  /**
   * Rewrite the slots/props if alteration is checked.
   *
   * @param array $configuration
   *   The plugin configuration.
   * @param mixed $value
   *
   * @return mixed|array
   */
  protected function alterField(array $configuration, mixed $value): mixed {
    if (empty($configuration['settings']['alter']['alter_field'])) {
      return $value;
    }
    // @todo test if we need to xss filter.
    return [
      '#type' => 'inline_template',
      '#template' => $configuration['settings']['alter']['text'],
      '#context' => [
        $this->propId => $value,
      ],
    ];
  }

  /**
   * Return the field items.
   *
   * @return \Drupal\Core\Field\FieldItemListInterface|null
   */
  protected function getFieldItems(): FieldItemListInterface|null {
    if (empty($this->configuration['settings']['ui_patterns_field_formatter'])) {
      return NULL;
    }

    return $this->configuration['settings']['ui_patterns_field_formatter']['items'] ?? NULL;
  }

  /**
   * Get the settings from the plugin configuration.
   *
   * @param array $setting
   *
   * @return mixed|null
   */
  protected function getSettingsFromConfiguration(array $setting) {
    $configuration = $this->getConfiguration();
    return NestedArray::getValue($configuration, $setting);
  }

}
