<?php

namespace Drupal\Tests\ui_patterns_field_formatters\Kernel;

use Drupal\Component\Plugin\PluginBase;
use Drupal\Core\Field\FieldConfigInterface;
use Drupal\Core\Plugin\Context\ContextDefinition;
use Drupal\Core\Plugin\Context\EntityContextDefinition;
use Drupal\ui_patterns_field_formatters\Plugin\Derivative\FieldPropertiesSourcePropDeriver;
use Drupal\ui_patterns_field_formatters\Plugin\Derivative\FieldPropertiesSourceSlotDeriver;
use Drupal\ui_patterns_field_formatters\Plugin\UiPatterns\Source\FieldPropertiesSourceProp;
use Drupal\ui_patterns_field_formatters\Plugin\UiPatterns\Source\FieldPropertiesSourceSlot;

/**
 * Tests the ui_patterns_field_formatters deriver plugin.
 *
 * @group ui_patterns_field_formatters
 *
 * @SuppressWarnings(PHPMD.CamelCaseParameterName)
 * @SuppressWarnings(PHPMD.CamelCaseVariableName)
 * @SuppressWarnings(PHPMD.LongClassName)
 * @SuppressWarnings(PHPMD.LongVariable)
 */
class PluginTest extends UIPatternsFieldFormattersTestBase {

  protected string $deriverID;

  /**
   * {@inheritdoc}
   */
  protected function setUp(): void {
    parent::setUp();
    $this->setUpMockDeriver();
  }

  /**
   * Set tup the deriver mock.
   *
   * @return void
   */
  private function setUpMockDeriver() {
    $field_storage_definition = $this->field->getFieldStorageDefinition();
    $field_property_definitions = $field_storage_definition->getPropertyDefinitions();
    $property_definition = $field_property_definitions[$this->property];
    $entity_type_labels = \Drupal::service('entity_type.repository')
      ->getEntityTypeLabels();
    $derivative_id = $this->entityType->getEntityType()->getBundleOf() . PluginBase::DERIVATIVE_SEPARATOR . $this->entityType->id() . PluginBase::DERIVATIVE_SEPARATOR . $this->field->getName() . PluginBase::DERIVATIVE_SEPARATOR . $this->property;
    $derivative['id'] = $derivative_id;
    $this->deriverID = $derivative_id;
    // label
    // description
    // prop_types
    // context_definitions.
    $derivative['label'] = sprintf('%s %s %s: %s',
      $entity_type_labels[$this->entityType->getEntityType()->getBundleOf()],
      $this->entityType->id(),
      // $e*/ntity_type_labels[$entity_type_id], // category
    // Admin label
      $this->field->getLabel(),
      $property_definition->getLabel(),
    );
    $derivative['description'] = sprintf('%s %s: property %s (%s) of field %s (%s)',
    // Category.
      $entity_type_labels[$this->entityType->getEntityType()->getBundleOf()],
    // Category.
      $this->entityType->id(),
      $property_definition->getLabel(),
      $this->property,
      $this->field->getLabel(),
      $this->field->getName()
    );
    // Add a dependency on the field if it is configurable.
    if ($this->field instanceof FieldConfigInterface) {
      $derivative['config_dependencies'][$this->field->getConfigDependencyKey()][] = $this->field->getConfigDependencyName();
    }
    $derivative['ui_patterns_field_formatter'] = [
      'entity_type' => $this->entityType->getEntityType()->getBundleOf(),
      'bundle' => $this->entityType->id(),
      'field_name' => $this->field->getName(),
      'field_name_parent' => '',
      'property' => $this->property,
      'entity_type_parent' => '',
      'main_property' => FALSE,
      'field_definition' => $this->field,
      'entity_reference' => FALSE,
    ];
    $derivative['context_definitions'] = 'ui_patterns_field_formatters';
    $derivative['provider'] = 'ui_patterns_field_formatters';
    $this->setUpContext($derivative, $entity_type_labels);
    $this->mockUIPatternFFExpectedDefinitions = $derivative;
  }

  /**
   * Callback data provider to test one slot, one prop.
   *
   * @return array[]
   */
  public function providerDeriverType(): array {
    return [
      [
        'deriver_type' => 'slot',
        'prop_types' => ['slot'],
        'deriver' => FieldPropertiesSourceSlotDeriver::class,
        'class' => FieldPropertiesSourceSlot::class,
      ],
      [
        'deriver_type' => 'prop',
        'prop_types' => ['string'],
        'deriver' => FieldPropertiesSourcePropDeriver::class,
        'class' => FieldPropertiesSourceProp::class,
      ],
    ];
  }

  /**
   * Set up the context definitions of the deriver.
   *
   * @param array $derivative
   * @param $entity_type_labels
   *
   * @return void
   */
  private function setUpContext(array &$derivative, $entity_type_labels) {
    $context_definition = EntityContextDefinition::fromEntityTypeId($this->entityType->getEntityType()
      ->getBundleOf())
      ->setLabel($entity_type_labels[$this->entityType->getEntityType()
        ->getBundleOf()]);
    // We explicitly set the bundle, but won't work if the context value only specifies the entity type.
    $context_definition->addConstraint('Bundle', [$this->entityType->id()]);

    // We explicitly set the field name, so when props are listed for a field, we only includes field properties
    //   for that field.
    $field_name_context = new ContextDefinition('string');
    $field_name_context->setRequired();
    $field_name_context->setDefaultValue($this->field->getName());
    $field_name_context->addConstraint('AllowedValues', [$this->field->getName()]);

    // We explicitly set the bundle, to ensure  when props are listed for a field and a bundle,
    //   we do not list properties from the same field name but from another bundle of the same entity type.
    $bundle_context = new ContextDefinition('string');
    $bundle_context->setRequired();
    $bundle_context->setDefaultValue($this->entityType->id());
    $bundle_context->addConstraint('AllowedValues', [$this->entityType->id()]);

    $derivative['context_definitions'] = [
      'entity' => $context_definition,
      'field_name' => $field_name_context,
      'bundle' => $bundle_context,
    ];
  }

  /**
   * Set the data provider inside the mock deriver.
   *
   * @param array $prop_types
   * @param string $deriver
   * @param string $class
   *
   * @return void
   */
  private function setDeriverType(array $prop_types, string $deriver, string $class) {
    $this->mockUIPatternFFExpectedDefinitions['prop_types'] = $prop_types;
    $this->mockUIPatternFFExpectedDefinitions['deriver'] = $deriver;
    $this->mockUIPatternFFExpectedDefinitions['class'] = $class;
  }

  /**
   * Tests getDefinitions() and getDefinition().
   *
   * @return void
   *
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   *
   * @dataProvider providerDeriverType
   */
  public function testDerivative(string $deriver_type, array $prop_types, string $deriver, string $class) {
    $this->setDeriverType($prop_types, $deriver, $class);
    // Ensure that getDefinitions() returns the expected definitions.
    $this->assertEquals($this->mockUIPatternFFExpectedDefinitions, $this->uiPatternsSourceManager->getDefinition('field_properties_source_' . $deriver_type . PluginBase::DERIVATIVE_SEPARATOR . $this->deriverID));
  }

}
