<?php

namespace Drupal\ui_patterns_legacy\Drush\Commands;

use Drupal\Component\Plugin\Discovery\CachedDiscoveryInterface;
use Drupal\ui_patterns_legacy\ComponentConverter;
use Drupal\ui_patterns_legacy\ComponentDiscovery;
use Drupal\ui_patterns_legacy\ComponentWriter;
use Drush\Attributes as CLI;
use Drush\Commands\DrushCommands;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\Finder\Finder;

/**
 * Drush commands of UI Patterns Legacy module.
 */
final class UiPatternsLegacyCommands extends DrushCommands {

  /**
   * {@inheritdoc}
   */
  public function __construct(
    private readonly ComponentConverter $converter,
    private readonly ComponentDiscovery $discovery,
    private readonly ComponentWriter $writer,
    private readonly CachedDiscoveryInterface $componentsManager
  ) {
    parent::__construct();
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('ui_patterns_legacy.component_converter'),
      $container->get('ui_patterns_legacy.discovery'),
      $container->get('ui_patterns_legacy.component_writer'),
      $container->get('plugin.manager.sdc')
    );
  }

  /**
   * Migrate components from UI Patterns 1.x to UI Patterns 2.x.
   */
  #[CLI\Command(name: 'ui-patterns:migrate', aliases: ['upm'])]
  #[CLI\Argument(name: 'extension', description: 'Module or theme machine name.')]
  #[CLI\Usage(name: 'ui-patterns:migrate my_theme', description: 'Migrate components from a theme.')]
  public function migrate(string $extension) {
    $legacy_definitions = $this->discovery->discover($extension);
    $extension_path = $this->discovery->getExtensionPath($extension);
    $this->converter->setExtension($extension);
    foreach ($legacy_definitions as $legacy_definition) {
      // First time to create all definitions.
      $component = $this->converter->convert($legacy_definition);
      // Second time for tasks which need all definitions to be already created.
      // @todo make it works better.
      $this->componentsManager->clearCachedDefinitions();
      $component = $this->converter->convert($legacy_definition);
      $component_id = $legacy_definition['id'];
      $this->writer->writeDefinition($component_id, $component, $extension_path);
      $this->writer->copyTemplates($extension_path, $legacy_definition);
      $this->writer->copyAssets($extension_path, $legacy_definition);
    }
    $components = $this->componentsManager->getDefinitions();
    foreach ($components as $component) {
      if ($extension !== $component['provider']) {
        continue;
      }
      $errors = $this->converter->validate($component);
      foreach ($errors as $error) {
        $this->io()->error($error);
      }
    }
  }

  /**
   * Upgrade code.
   */
  #[CLI\Command(name: 'ui-patterns:upgrade', aliases: ['upu'])]
  #[CLI\Argument(name: 'extension', description: 'Module or theme machine name.')]
  #[CLI\Usage(name: 'ui-patterns:upgrade my_theme', description: 'Upgrade theme code.')]
  public function upgrade(string $extension) {
    $extension_path = $this->discovery->getExtensionPath($extension);
    $this->replaceDeprecatedCalls($extension_path);
    $this->changeInfoFile($extension, $extension_path);
    // @todo Update composer.json?
  }

  /**
   * Replace deprecated calls.
   *
   * LinksSettingType::normalize() is now automatically managed with
   * PropTypeInterface::normalize(). No need for manual calls anymore.
   * So let's replace them by a temporary dummy.
   */
  protected function replaceDeprecatedCalls(string $extension_path): void {
    $finder = new Finder();
    $patterns = ['*.php', '*.inc', '*.module', '*.theme'];
    $finder->files()->name($patterns)->in($extension_path);
    foreach ($finder as $file) {
      $path = $file->getPathname();
      $content = file_get_contents($path);
      $content = str_replace('Drupal\ui_patterns_settings\Plugin\UiPatterns\SettingType', 'Drupal\ui_patterns_legacy\Dummy', $content);
      file_put_contents($path, $content);
    }
  }

  /**
   * Change info file.
   */
  protected function changeInfoFile(string $extension, string $extension_path): void {
    $info_path = $extension_path . "/" . $extension . ".info.yml";
    $content = file_get_contents($info_path);
    $content = preg_replace('/ui_patterns_settings.*/', 'ui_patterns:ui_patterns_legacy', $content);
    file_put_contents($info_path, $content);
  }

}
