<?php

declare(strict_types=1);

namespace Drupal\ui_patterns\Element;

use Drupal\Core\Security\TrustedCallbackInterface;
use Drupal\ui_patterns\Plugin\UiPatterns\PropType\SlotPropType;

/**
 * Our additions to the SDC render element.
 */
class ComponentElementAlter implements TrustedCallbackInterface {

  /**
   * {@inheritdoc}
   */
  public static function trustedCallbacks() {
    return ['alter'];
  }

  /**
   * Alter SDC component element.
   */
  public function alter(array $element) {
    foreach ($element["#slots"] as $slot_id => $slot) {
      // Because SDC validator is sometimes confused by a null slot.
      if (is_null($slot)) {
        unset($element['#slots'][$slot_id]);
        continue;
      }
      // Because SDC validator is sometimes confused by an empty slot.
      if (is_array($slot) && empty($slot)) {
        unset($element['#slots'][$slot_id]);
        continue;
      }
      $element["#slots"][$slot_id] = SlotPropType::normalize($slot);
    }
    $component = \Drupal::service("plugin.manager.sdc")->find($element['#component']);
    $props = $component->metadata->schema['properties'] ?? [];
    foreach ($element["#props"] as $prop_id => $prop) {
      if (!isset($props[$prop_id])) {
        continue;
      }
      $prop_type = $props[$prop_id]['ui_patterns']['type_definition'];
      $element["#props"][$prop_id] = $prop_type->normalize($prop);
    }
    return $element;
  }

  /**
   * Process stories slots.
   *
   * Stories slots have no "#" prefix in render arrays. Let's add them.
   * A bit like UI Patterns 1.x's PatternPreview::getPreviewMarkup()
   * This method belongs here because used by both ui_patterns_library and
   * ui_patterns_legacy.
   */
  public function processStoriesSlots(array $slots): array {
    foreach ($slots as $slot_id => $slot) {
      if (!is_array($slot)) {
        continue;
      }
      if (array_is_list($slot)) {
        $slots[$slot_id] = $this->processStoriesSlots($slot);
      }
      $slot_keys = array_keys($slot);
      $render_keys = ["theme", "type", "markup", "plain_text"];
      if (count(array_intersect($slot_keys, $render_keys)) > 0) {
        foreach ($slot as $key => $value) {
          if (is_array($value)) {
            $value = $this->processStoriesSlots($value);
          }
          if (str_starts_with($key, "#")) {
            continue;
          }
          $slots[$slot_id]["#" . $key] = $value;
          unset($slots[$slot_id][$key]);
        }
      }
    }
    return $slots;
  }

}
