<?php

declare(strict_types=1);

namespace Drupal\ui_patterns\Element;

use Drupal\Core\Plugin\Component;
use Drupal\Core\Security\TrustedCallbackInterface;
use Drupal\Core\Theme\ComponentPluginManager;
use Drupal\ui_patterns\PropTypeAdapterPluginManager;
use Drupal\ui_patterns\SourcePluginManager;

/**
 * Component render element builder.
 */
class ComponentElementBuilder implements TrustedCallbackInterface {

  /**
   * {@inheritdoc}
   */
  public static function trustedCallbacks() {
    return ['build'];
  }

  /**
   * Constructs a ComponentElementBuilder.
   */
  public function __construct(protected SourcePluginManager $sourcesManager, protected PropTypeAdapterPluginManager $adaptersManager, protected ComponentPluginManager $componentPluginManager) {
  }

  /**
   * Build component data provided to the SDC element.
   */
  public function build(array $element): array {
    if (!isset($element['#ui_patterns'])) {
      return $element;
    }
    $configuration = $element['#ui_patterns'];
    $contexts = $element['#source_contexts'] ?? [];
    $component = $this->componentPluginManager->find($element['#component']);
    $element = $this->buildProps($element, $component, $configuration, $contexts);
    $element = $this->buildSlots($element, $component, $configuration, $contexts);
    $element['#propsAlter'] = [];
    return $element;
  }

  /**
   * Add props to the renderable.
   */
  protected function buildProps(array $build, Component $component, array $configuration, array $contexts): array {
    $props = $component->metadata->schema['properties'] ?? [];
    foreach ($props as $prop_id => $prop_definition) {
      $prop_configuration = $configuration['props'][$prop_id] ?? [];
      $build = $this->buildProp($build, $prop_id, $prop_definition, $prop_configuration, $contexts);
    }
    if (isset($configuration["variant_id"])) {
      $build['#props']["variant"] = $configuration["variant_id"];
    }
    return $build;
  }

  /**
   * Add a single prop to the renderable.
   */
  protected function buildProp(array $build, string $prop_id, array $definition, array $configuration, array $source_contexts): array {
    if (isset($build["#props"][$prop_id])) {
      // Keep existing props. No known use case yet.
      return $build;
    }
    $prop_type = $definition['ui_patterns']['type_definition'];
    $source_id = array_key_exists("source_id", $configuration) ? $configuration["source_id"] : $this->sourcesManager->getPropTypeDefault($prop_type->getPluginId(), $source_contexts);
    if (!$source_id) {
      return $build;
    }
    $source = $this->sourcesManager->createInstance(
      $source_id,
      SourcePluginManager::buildPluginConfiguration($prop_id, $definition, $configuration, $source_contexts)
    );
    if (!$source) {
      return $build;
    }
    $build = $source->alterComponent($build);
    $data = $source->getPropValue();
    if (isset($definition['ui_patterns']['prop_type_adapter'])) {
      $prop_type_adapter_id = $definition['ui_patterns']['prop_type_adapter'];
      $prop_type_adapter = $this->adaptersManager->createInstance($prop_type_adapter_id);
      $data = $prop_type_adapter->transform($data);
    }
    if (empty($data) && $prop_type->getPluginId() !== 'attributes') {
      // For JSON Schema validator, empty value is not the same as missing
      // value, and we want to prevent some of the prop types rules to be
      // applied on empty values: string pattern, string format, enum, number
      // min/max...
      // However, we don't remove empty attributes to avoid an error with
      // Drupal\Core\Template\TwigExtension::createAttribute() when themers
      // forget to use the default({}) filter.
      return $build;
    }
    $build["#props"][$prop_id] = $data;
    return $build;
  }

  /**
   * Add slots to the renderable.
   */
  protected function buildSlots(array $build, Component $component, array $configuration, array $contexts): array {
    $slots = $component->metadata->slots ?? [];
    foreach ($slots as $slot_id => $slot_definition) {
      $slot_configuration = $configuration['slots'][$slot_id] ?? [];
      $build = $this->buildSlot($build, $slot_id, $slot_definition, $slot_configuration, $contexts);
    }
    return $build;
  }

  /**
   * Add a single slot to the renderable.
   */
  protected function buildSlot(array $build, string $slot_id, array $definition, array $configuration, array $contexts): array {
    if (isset($build["#slots"][$slot_id])) {
      // Keep existing slots. Used by ComponentLayout for example.
      return $build;
    }
    if (!isset($configuration["sources"])) {
      return $build;
    }
    // Slots can have many sources while props can have only one.
    $build["#props"][$slot_id] = [];
    foreach ($configuration["sources"] as $source_configuration) {
      $build = $this->addSlotSource($build, $slot_id, $definition, $source_configuration, $contexts);
    }
    if (count($build["#slots"][$slot_id]) === 1) {
      $build["#slots"][$slot_id] = $build["#slots"][$slot_id][0];
    }
    return $build;
  }

  /**
   * Add source data to the slot.
   */
  protected function addSlotSource(array $build, string $slot_id, array $definition, array $configuration, array $source_contexts): array {
    $source_id = $configuration['source_id'];
    if (!$this->sourcesManager->isApplicable($source_id, $source_contexts)) {
      return $build;
    }
    $source = $this->sourcesManager->createInstance(
        $source_id,
        SourcePluginManager::buildPluginConfiguration($slot_id, $definition, $configuration, $source_contexts)
      );
    $build = $source->alterComponent($build);
    $build["#slots"][$slot_id][] = $source->getPropValue();
    return $build;
  }

}
