<?php

namespace Drupal\ui_patterns\Element;

use Drupal\Component\Utility\NestedArray;
use Drupal\Core\Form\FormStateInterface;

/**
 * Provides a Component form builder element.
 *
 * Use #component_id to set an unchangeable component.
 *
 * The value of the form element contains following keys:
 *  [
 *    'component_id' => 'my_module:my_component',
 *    'variant_id' => 'variant',
 *    'slots' => [
 *      'slots_id' => [
 *        ['source_id' => 'id', 'value' => 'Source value']
 *      ],
 *    ],
 *    'props' => [
 *      ['props_id' =>
 *       ['source_id' => 'id', 'value' => 'Source value ']
 *      ]
 *    ],
 *  ]
 * The same component value is returned in the form element.
 *
 * Usage example:
 *
 * @code
 * $form['component_form'] = [
 *   '#type' => 'component_form',
 *
 *   '#default_value' => [
 *     'component_id' => 'my_module:my_component',
 *     'variant_id' => 'variant',
 *     'slots' => [],
 *     'props' => [],
 *   ],
 * ];
 * @endcode
 *
 * @FormElement("component_form")
 */
class ComponentForm extends ComponentFormBase {

  /**
   * {@inheritdoc}
   */
  public function getInfo() {
    $class = get_class($this);
    return [
      // The component id.
      '#component_id' => NULL,
      '#input' => TRUE,
      '#multiple' => FALSE,
      '#default_value' => NULL,
      '#source_contexts' => [],
      '#process' => [
        [$class, 'buildForm'],
      ],
      '#element_validate' => [
        [$class, 'validateFormElement'],
      ],
      '#theme_wrappers' => ['form_element'],
    ];
  }

  /**
   * {@inheritdoc}
   */
  public static function valueCallback(&$element, $input, FormStateInterface $form_state) {
    if ($input) {
      $value = [
        'component_id' => $input['component_id'] ?? NULL,
        'variant_id' => $input['component_form']['variant_id'] ?? NULL,
        'props' => $input['component_form']['props'] ?? [],
        'slots' => $input['component_form']['slots'] ?? [],
      ];
      $element['#default_value'] = $value;
      return $value;
    }
    else {
      return [
        'component_id' => NULL,
        'variant_id' => NULL,
        'props' => [],
        'slots' => [],
      ];
    }

  }

  /**
   * Processes the main form element including component selector.
   */
  public static function buildForm(array &$element, FormStateInterface $form_state) {
    $trigger_element = $form_state->getTriggeringElement();
    if ($form_state->isRebuilding() && isset($trigger_element['#ui_patterns'])) {
      $value = $form_state->getValue(self::getComponentFormStateParents($trigger_element['#parents']));
      self::valueCallback($element, $value, $form_state);
    }

    $initial_component_id = $element['#component_id'] ?? NULL;
    $component_id = $initial_component_id ?? $element['#default_value']['component_id'] ?? NULL;
    $wrapper_id = 'ui-patterns-component';
    if ($initial_component_id === NULL) {
      $element["component_id"] = self::expandAjax(self::buildComponentSelectorForm(
        $wrapper_id,
        $component_id
      ));
    }
    $element["component_form"] = self::buildComponentForm(
      $element,
      $wrapper_id,
      $component_id
    );
    $element['#tree'] = TRUE;
    return $element;
  }

  /**
   * Processes the selected component from element.
   *
   * @return array
   *   The component form.
   */
  private static function buildComponentForm(
    array $element,
    string $wrapper_id,
    ?string $component_id
  ): array {
    $form = [
      '#type' => 'container',
      '#prefix' => '<div id="' . $wrapper_id . '">',
      '#suffix' => '</div>',
    ];
    if (!$component_id) {
      return $form;
    }
    $form['variant_id'] = self::buildComponentVariantSelectorForm(
      $element,
      $element['#default_value']['variant_id'] ?? NULL,
    );
    $form['slots'] = self::buildSlotsForm($element, $component_id);
    $form['props'] = self::buildPropsForm($element, $component_id);
    return $form;
  }

  /**
   * Build components selector widget.
   *
   * @return array
   *   The component select.
   */
  private static function buildComponentSelectorForm(
    ?string $wrapper_id,
    ?string $selected_component_id
  ): array {
    $definition_groups = \Drupal::service("plugin.manager.sdc")->getGroupedDefinitions();
    $options = [];
    foreach ($definition_groups as $definition_group_id => $definition_group) {
      $group_options = [];
      foreach ($definition_group as $definition_id => $definition) {
        $group_options[$definition_id] = $definition['name'];
      }
      $options[$definition_group_id] = $group_options;
    }
    return [
      "#type" => "select",
      "#title" => t("Component"),
      "#options" => $options,
      '#default_value' => $selected_component_id,
      '#ajax' => [
        'callback' => [self::class, 'changeSelectorFormChangeAjax'],
        'wrapper' => $wrapper_id,
        'effect' => 'fade',
      ],
      '#empty_value' => '',
      '#empty_option' => t('- None -'),
      '#required' => TRUE,
    ];
  }

  /**
   * Build the variant select widget.
   *
   * @return array
   *   The variant select.
   */
  private static function buildComponentVariantSelectorForm(
    array $element,
    string|NULL $default_variant_id
  ): array {
    $component = self::getComponent($element);
    $definition = $component->getPluginDefinition();
    if (!isset($definition["variants"])) {
      return [];
    }
    $options = [];
    foreach ($definition["variants"] as $variant_id => $variant) {
      $options[$variant_id] = $variant["title"] ?? $variant_id;
    }
    return [
      "#type" => "select",
      "#title" => t("Variant"),
      "#options" => $options,
      '#default_value' => $default_variant_id,
    ];
  }

  /**
   * Build slots form.
   */
  private static function buildSlotsForm(array $element, string $component_id): array {
    return [
      '#title' => 'Slots',
      '#type' => 'component_slots_form',
      '#component_id' => $component_id,
      '#source_contexts' => $element['#source_contexts'],
      '#ajax_url' => $element['#ajax_url'],
      '#access' => $element['#render_slots'] ?? TRUE,
      '#default_value' => [
        'slots' => $element['#default_value']['slots'],
      ],
    ];
  }

  /**
   * Build props form.
   */
  private static function buildPropsForm(array $element, string $component_id): array {
    return [
      '#title' => 'Props',
      '#type' => 'component_props_form',
      '#component_id' => $component_id,
      '#source_contexts' => $element['#source_contexts'],
      '#ajax_url' => $element['#ajax_url'],
      '#access' => $element['#render_props'] ?? TRUE,
      '#default_value' => [
        'props' => $element['#default_value']['props'],
      ],
    ];
  }

  /**
   * Ajax callback for component selector change.
   */
  public static function changeSelectorFormChangeAjax(
    array $form,
    FormStateInterface $form_state
  ) {
    $parents = $form_state->getTriggeringElement()['#array_parents'];
    $sub_form = NestedArray::getValue($form, array_slice($parents, 0, -1));
    $form_state->setRebuild();
    return $sub_form['component_form'];
  }

  /**
   * Form element validation handler.
   */
  public static function validateFormElement(array &$element, FormStateInterface $form_state) {
    $trigger_element = $form_state->getTriggeringElement();
    if (isset($trigger_element['#ui_patterns']) === FALSE) {
      $form_state->setValueForElement($element, $element['#value']);
    }
  }

}
