<?php

namespace Drupal\ui_patterns\Element;

use Drupal\Component\Utility\Html;
use Drupal\Component\Utility\NestedArray;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Render\Element;
use Drupal\ui_patterns\SourcePluginManager;

/**
 * Provides a Component form builder element.
 *
 * Use #component_id to set an unchangeable component.
 *
 * The value of the form element contains following keys:
 *  [
 *    'slots' => [
 *      'slots_id' => [
 *        ['source_id' => 'id', 'value' => 'Source value']
 *      ],
 *    ],
 *  ]
 * The same component value is returned in the form element.
 *
 * Usage example:
 *
 * @code
 * $form['slots'] = [
 *   '#type' => 'component_slots_form',
 *   '#default_value' => [
 *     'slots' => [],
 *   ],
 * ];
 * @endcode
 *
 * @FormElement("component_slots_form")
 */
class ComponentSlotsForm extends ComponentFormBase {

  /**
   * {@inheritdoc}
   */
  public function getInfo() {
    $class = get_class($this);
    return [
      '#input' => TRUE,
      '#multiple' => FALSE,
      '#default_value' => NULL,
      '#source_contexts' => [],
      '#process' => [
        [$class, 'buildForm'],
      ],
      '#theme_wrappers' => ['container'],
    ];
  }

  /**
   * Processes slotes form element.
   */
  public static function buildForm(array &$element, FormStateInterface $form_state): array {

    /** @var \Drupal\ui_patterns\SourcePluginManager $sources_manager */
    $sources_manager = \Drupal::service("plugin.manager.ui_patterns_source");
    $source_contexts = $element['#source_contexts'] ?? [];
    $sources = $sources_manager->getDefinitionsForPropType('slot', $source_contexts);
    $source_ids = array_keys($sources);
    $source_ids = array_combine($source_ids, $source_ids);
    $valid_source_plugins = $sources_manager->createInstances(
      $source_ids,
      SourcePluginManager::buildPluginConfiguration('slot', [], [], $source_contexts),
    );
    $options = [];
    foreach ($valid_source_plugins as $valid_source_plugin) {
      $options[$valid_source_plugin->getPluginId()] = $valid_source_plugin->label();
    }
    $component = self::getComponent($element);
    if (!isset($component->metadata->slots) || count(
        $component->metadata->slots
      ) === 0) {
      hide($element);
      return $element;
    }
    $configuration = $element['#default_value']['slots'] ?? [];
    foreach ($component->metadata->slots as $slot_id => $slot) {
      $element[$slot_id] = self::buildSlotForm($form_state, $slot_id, $slot, $configuration[$slot_id] ?? [], $options, $source_contexts);
    }
    return $element;
  }

  /**
   * Build single slot form.
   */
  protected static function buildSlotForm(FormStateInterface $form_state, string $slot_id, array $definition, array $configuration, array $options, array $source_contexts): array {
    $element = [];
    $wrapper_id = 'ui-patterns-slot-' . $slot_id;
    $element['#slot_id'] = $slot_id;
    $element['sources'] = self::buildSourcesForm($form_state, $slot_id, $definition, $configuration, $wrapper_id, $source_contexts);
    $element['add_more_button'] = self::buildSourceSelector($slot_id, $wrapper_id, $options);
    return $element;
  }

  /**
   * Build single slot's sources form.
   */
  protected static function buildSourcesForm(FormStateInterface $form_state, string $slot_id, array $definition, array $configuration, string $wrapper_id, array $source_contexts): array {
    $element = [
      '#theme' => 'field_multiple_value_form',
      '#title' => $definition['title'],
      '#cardinality_multiple' => TRUE,
      '#prefix' => '<div id="' . $wrapper_id . '">',
      '#suffix' => '</div>',
    ];
    // Add fake #field_name to avoid errors from
    // template_preprocess_field_multiple_value_form.
    $element['#field_name'] = "foo";
    if (!isset($configuration['sources'])) {
      return $element;
    }
    foreach ($configuration['sources'] as $delta => $source_configuration) {
      if (!isset($source_configuration['source_id'])) {
        continue;
      }
      $element[$delta] = self::buildSourceForm($form_state, $slot_id, $definition, $source_configuration, $delta, $wrapper_id, $source_contexts);
    }
    return $element;
  }

  /**
   * Build single source form.
   */
  protected static function buildSourceForm(FormStateInterface $form_state, string $slot_id, array $definition, array $configuration, int $delta, string $wrapper_id, array $source_contexts): array {
    $element = [];
    $sources_manager = \Drupal::service("plugin.manager.ui_patterns_source");
    $source = $sources_manager->createInstance(
        $configuration['source_id'],
        SourcePluginManager::buildPluginConfiguration($slot_id, $definition, $configuration, $source_contexts)
      );
    $element['source'] = $source->settingsForm([], $form_state);
    $element['source_id'] = [
      '#type' => 'hidden',
      '#value' => $source->getPluginId(),
    ];
    $element['_weight'] = [
      '#type' => 'weight',
      '#title' => t(
          'Weight for row @number',
          ['@number' => $delta + 1]
      ),
      '#title_display' => 'invisible',
      '#delta' => count($element),
      '#default_value' => $configuration['_weight'] ?? $delta,
      '#weight' => 100,
    ];
    $element['_remove'] = self::buildRemoveSourceButton($slot_id, $wrapper_id, $delta);
    return $element;
  }

  /**
   * Build widget to remove source.
   */
  protected static function buildRemoveSourceButton(string $slot_id, string $wrapper_id, int $delta): array {
    $delete_action = [
      '#type' => 'submit',
      '#name' => strtr($slot_id, '-', '_') . $delta . '_remove',
      '#value' => t('Delete'),
      '#submit' => [self::class . '::removeSlotItemSubmitAjax'],
      '#access' => TRUE,
      '#delta' => $delta,
      '#ui_patterns' => [
        'parents' => 5,
      ],
      '#slot_id' => $slot_id,
      '#ajax' => [
        'callback' => [self::class, 'removeSlotItemAjax'],
        'wrapper' => $wrapper_id,
        'effect' => 'fade',
      ],
    ];
    return [
      '#type' => 'ui_patterns_actions',
      '#ui_patterns_header' => TRUE,
      'dropdown_actions' => [
        self::expandComponentButton($delete_action),
      ],
    ];
  }

  /**
   * Build source selector.
   */
  protected static function buildSourceSelector(string $slot_id, string $wrapper_id, array $options): array {
    $action_buttons = [];
    foreach ($options as $source_id => $source_label) {
      $action_buttons[$source_id] = self::expandComponentButton([
        '#type' => 'submit',
        '#name' => strtr($slot_id, '-', '_') . $source_id . '_add_more',
        '#value' => t('Add %source', ['%source' => $source_label]),
        '#submit' => [
          self::class . '::addSlotItemSubmitAjax',
        ],
        '#access' => TRUE,
        '#ui_patterns' => TRUE,
        '#slot_id' => $slot_id,
        '#source_id' => $source_id,
        '#ajax' => [
          'callback' => [
            self::class,
            'addSlotItemAjax',
          ],
          'wrapper' => $wrapper_id,
          'effect' => 'fade',
        ],
      ]);
    }
    return self::buildComponentDropbutton($action_buttons);
  }

  /**
   * Ajax submit handler: Add slot item.
   */
  public static function addSlotItemSubmitAjax(array $form, FormStateInterface $form_state) {
    $trigger_element = $form_state->getTriggeringElement();
    $slot_id = $trigger_element['#slot_id'];
    $source_id = $trigger_element['#source_id'];
    $component_form_parents = self::getComponentFormStateParents($trigger_element['#parents']);
    $configuration = $form_state->getValue($component_form_parents);
    $configuration['component_form']['slots'][$slot_id]['sources'][]['source_id'] = $source_id;
    $form_state->setValue($component_form_parents, $configuration);
    $form_state->setRebuild();
  }

  /**
   * Ajax handler: Add slot item.
   */
  public static function addSlotItemAjax(array $form, FormStateInterface $form_state) {
    $parents = $form_state->getTriggeringElement()['#array_parents'];
    $form = NestedArray::getValue($form, array_slice($parents, 0, -2));
    return $form['sources'];
  }

  /**
   * Ajax submit handler: Remove slot item.
   */
  public static function removeSlotItemSubmitAjax(array $form, FormStateInterface $form_state): void {
    $trigger_element = $form_state->getTriggeringElement();
    $delta = $trigger_element['#delta'];
    $slot_id = $trigger_element['#slot_id'];
    $component_form_parents = self::getComponentFormStateParents($trigger_element['#parents']);
    $configuration = $form_state->getValue($component_form_parents);
    unset($configuration['component_form']['slots'][$slot_id]['sources'][$delta]);
    $form_state->setValue($component_form_parents, $configuration);
    $form_state->setRebuild();
  }

  /**
   * Ajax handler: Remove slot item.
   */
  final public static function removeSlotItemAjax(array $form, FormStateInterface $form_state) {
    $parents = $form_state->getTriggeringElement()['#array_parents'];
    return NestedArray::getValue($form, array_slice($parents, 0, -4));
  }

  /**
   * Build drop button.
   *
   * @param array $elements
   *   Elements for drop button.
   *
   * @return array
   *   Drop button array.
   */
  protected static function buildComponentDropbutton(array $elements = []): array {
    $build = [
      '#type' => 'container',
      '#attributes' => ['class' => ['ui-patterns-dropbutton-wrapper']],
    ];

    $operations = [];
    // Because we are cloning the elements into title sub element we need to
    // sort children first.
    foreach (Element::children($elements, TRUE) as $child) {
      // Clone the element as an operation.
      $operations[$child] = ['title' => $elements[$child]];

      // Flag the original element as printed so it doesn't render twice.
      $elements[$child]['#printed'] = TRUE;
    }

    $build['operations'] = [
      '#type' => 'ui_patterns_operations',
      // Even though operations are run through the "links" element type, the
      // theme system will render any render array passed as a link "title".
      '#links' => $operations,
      '#dropbutton_type' => 'small',
    ];

    return $build + $elements;
  }

  /**
   * Expand button base array into a paragraph widget action button.
   *
   * @param array $button_base
   *   Button base render array.
   *
   * @return array
   *   Button render array.
   */
  protected static function expandComponentButton(array $button_base): array {
    // Do not expand elements that do not have submit handler.
    if (empty($button_base['#submit'])) {
      return $button_base;
    }

    $button = $button_base + [
      '#type' => 'submit',
      '#theme_wrappers' => ['input__submit__ui_patterns_action'],
    ];

    // Html::getId will give us '-' char in name but we want '_' for now so
    // we use strtr to search&replace '-' to '_'.
    $button['#name'] = strtr(Html::getId($button_base['#name']), '-', '_');
    $button['#id'] = Html::getUniqueId($button['#name']);

    if (isset($button['#ajax'])) {
      $button['#ajax'] += [
        'effect' => 'fade',
        // Since a normal throbber is added inline, this has the potential to
        // break a layout if the button is located in dropbuttons. Instead,
        // it's safer to just show the fullscreen progress element instead.
        'progress' => ['type' => 'fullscreen'],
      ];
    }

    return self::expandAjax($button);
  }

}
