<?php

declare(strict_types=1);

namespace Drupal\ui_patterns\Form;

use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Form\SubformState;
use Drupal\Core\Url;

/**
 * Component form builder trait.
 */
trait ComponentFormBuilderTrait {

  /**
   * Adapter function to get plugin configuration.
   *
   * Overwrite to return settings/options of the
   * current plugin.
   *
   * @return array
   *   The plugin settings/options.
   */
  protected function getComponentConfiguration(): array {
    return $this->configuration['ui_patterns'] ?? [];
  }

  /**
   * Adapter function to set plugin configuration.
   *
   * Overwrite to return settings/options of the
   * current plugin.
   *
   * @param mixed $configuration
   *   The configuration to store.
   */
  protected function setComponentConfiguration($configuration): void {
    $this->configuration['ui_patterns'] = $configuration;
  }

  /**
   * Get component form default.
   *
   * To use with:
   * - PluginSettingsInterface::defaultSettings
   * - ConfigurableInterface::defaultConfiguration
   * - views/PluginBase::setOptionDefaults
   * - ...
   */
  public static function getComponentFormDefault() {
    return [
      "ui_patterns" => [
        "component_id" => NULL,
        "variant_id" => NULL,
        "slots" => [],
        "props" => [],
      ],
    ];
  }

  /**
   * Returns the ajax url.
   *
   * Some integrations plugins like the views plugin provide there
   * ajax urls for ajax form interaction. To support
   * those plugins overwrite this function.
   *
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The form state.
   *
   * @return \Drupal\Core\Url|null
   *   The ajax url.
   */
  protected function getAjaxUrl(FormStateInterface $form_state): ?Url {
    return NULL;
  }

  /**
   * Build the complete form.
   *
   * The form contains.
   *
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The form state.
   * @param array $source_contexts
   *   Source contexts.
   * @param string|null $initial_component_id
   *   The initial_component_id. If provided the component is changeable.
   * @param bool $render_slots
   *   TRUE if slots are editable.
   * @param bool $render_props
   *   TRUE if props are editable.
   */
  protected function buildComponentsForm(
    FormStateInterface $form_state,
    array $source_contexts = [],
    string $initial_component_id = NULL,
    bool $render_slots = TRUE,
    bool $render_props = TRUE,
  ): array {
    $form_state = $form_state instanceof SubformState ? $form_state->getCompleteFormState() : $form_state;
    return [
      'ui_patterns' => [
        '#type' => 'component_form',
        '#component_id' => $initial_component_id,
        '#ajax_url' => $this->getAjaxUrl($form_state),
        '#source_contexts' => $source_contexts,
        '#default_value' => $this->getComponentConfiguration(),
        '#render_slots' => $render_slots,
        '#render_props' => $render_props,
      ],
    ];
  }

  /**
   * Submit the component form.
   *
   * The form contains.
   *
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The form state.
   */
  public function submitComponentsForm(FormStateInterface $form_state) {
    $this->setComponentConfiguration($form_state->getValue('ui_patterns'));
  }

  /**
   * Build component renderable (a SDC render element).
   */
  public function buildComponentRenderable($component_id = NULL, $source_contexts = []) {
    return [
      '#type' => 'component',
      '#component' => $component_id ?? $this->getComponentConfiguration()['component_id'],
      '#ui_patterns' => $this->getComponentConfiguration(),
      '#source_contexts' => $source_contexts,
    ];
  }

}
