<?php

declare(strict_types=1);

namespace Drupal\ui_patterns\Plugin\UiPatterns\PropType;

use Drupal\Core\Plugin\ContainerFactoryPluginInterface;
use Drupal\Core\Url;
use Drupal\ui_patterns\PropTypePluginBase;
use Drupal\ui_patterns\SchemaManager\ReferencesResolver;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Provides a 'links' PropType.
 *
 * @PropType(
 *   id = "links",
 *   label = @Translation("Links"),
 *   description = @Translation("A list of link objects."),
 *   priority = 10,
 *   default_source = "menu",
 *   schema = {
 *     "type": "array",
 *     "items": {
 *       "type": "object",
 *       "properties": {
 *         "title": {"type": "string"},
 *         "url": { "$ref": "ui-patterns://url" },
 *         "attributes": { "$ref": "ui-patterns://attributes" },
 *         "link_attributes": { "$ref": "ui-patterns://attributes" },
 *         "below": { "type": "array", "items": { "type": "object" } }
 *       },
 *       "required": {"title"}
 *     }
 *   }
 * )
 */
class LinksPropType extends PropTypePluginBase implements ContainerFactoryPluginInterface {

  /**
   * {@inheritdoc}
   */
  public function __construct(array $configuration, $plugin_id, $plugin_definition, protected ReferencesResolver $referenceResolver) {
    parent::__construct($configuration, $plugin_id, $plugin_definition);
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    return new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $container->get('ui_patterns.schema_reference_solver')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function getSchema(): array {
    // Resolve prop type schema, because the resolver run by
    // JsonSchema\Validator (so, by SDC ComponentValidator) is not
    // recursive.
    $schema = $this->pluginDefinition['schema'];
    $schema = $this->referenceResolver->resolve($schema);
    return (array) $schema;
  }

  /**
   * {@inheritdoc}
   */
  public static function normalize(mixed $value): mixed {
    if (!is_array($value)) {
      return [];
    }
    return self::normalizeLinks($value);
  }

  /**
   * Normalize links.
   */
  protected static function normalizeLinks(array $value): array {
    // Don't inject URL object into patterns templates, use "title" as item
    // label and "url" as item target.
    foreach ($value as $index => &$item) {
      if (!is_array($item)) {
        unset($value[$index]);
        continue;
      }
      $item = self::normalizeAttributes($item);
      if (array_key_exists("text", $item)) {
        // Examples: links.html.twig, breadcrumb.html.twig, pager.html.twig,
        // views_mini_pager.html.twig.
        $item["title"] = $item["text"];
        unset($item["text"]);
      }
      if (!array_key_exists("title", $item)) {
        $item["title"] = $index;
      }
      if (array_key_exists("href", $item)) {
        // Examples: pager.html.twig, views_mini_pager.html.twig.
        $item["url"] = $item["href"];
        unset($item["href"]);
      }
      if (!isset($item["url"]) && isset($item["link"])) {
        // Example: links.html.twig.
        $item["url"] = $item["link"]["#url"];
        $item["url"]->setOptions($item["link"]["#options"]);
        unset($item["link"]);
      }
      $item = self::normalizeUrl($item);
      if (array_key_exists("below", $item)) {
        $item["below"] = self::normalize($item["below"]);
      }
    }
    return array_values($value);
  }

  /**
   * Normaize attributes in an item.
   */
  private static function normalizeAttributes(array $item): array {
    if (array_key_exists("attributes", $item) && is_a($item["attributes"], '\Drupal\Core\Template\Attribute')) {
      // In UI Patterns Settings, we kept the Attribute object here. It is not
      // possible anymore because SDC will not validate it against the prop
      // type schema.
      $item["attributes"] = $item["attributes"]->toArray();
      foreach ($item["attributes"] as $attr => $value) {
        if ($value instanceof Url) {
          $item["attributes"][$attr] = $value->toString();
        }
      }
    }
    return $item;
  }

  /**
   * Normaize URL in an item.
   *
   * Useful for: menu.html.twig, links.html.twig.
   */
  private static function normalizeUrl(array $item): array {
    if (!array_key_exists("url", $item)) {
      return $item;
    }
    $url = $item["url"];
    if (!($url instanceof Url)) {
      return $item;
    }
    if ($url->isRouted() && ($url->getRouteName() === '<nolink>')) {
      unset($item["url"]);
    }
    elseif ($url->isRouted() && ($url->getRouteName() === '<button>')) {
      unset($item["url"]);
    }
    else {
      $item["url"] = $url->toString();
    }
    $item = self::extractUrlOptions($url, $item);
    return $item;
  }

  /**
   * Extract URL options.
   */
  private static function extractUrlOptions(Url $url, array $item): array {
    $options = $url->getOptions();
    if (isset($options["attributes"])) {
      // In UI Patterns Settings, we built an Attribute object here. It is not
      // possible anymore because SDC will not validate it against the prop
      // type schema.
      $item["link_attributes"] = $options["attributes"];
      foreach ($item["link_attributes"] as $attr => $value) {
        if ($value instanceof Url) {
          $item["link_attributes"][$attr] = $value->toString();
        }
      }
    }
    return $item;
  }

}
