<?php

namespace Drupal\ui_patterns;

use Drupal\Component\Plugin\FallbackPluginManagerInterface;
use Drupal\Core\Cache\CacheBackendInterface;
use Drupal\Core\Extension\ModuleHandlerInterface;
use Drupal\Core\Plugin\DefaultPluginManager;

/**
 * PropType plugin manager.
 */
class PropTypePluginManager extends DefaultPluginManager implements FallbackPluginManagerInterface, SchemaGuesserInterface {

  /**
   * Constructs PropTypePluginManager object.
   *
   * @param \Traversable $namespaces
   *   An object that implements \Traversable which contains the root paths
   *   keyed by the corresponding namespace to look for plugin implementations.
   * @param \Drupal\Core\Cache\CacheBackendInterface $cache_backend
   *   Cache backend instance to use.
   * @param \Drupal\Core\Extension\ModuleHandlerInterface $module_handler
   *   The module handler to invoke the alter hook with.
   */
  public function __construct(\Traversable $namespaces, CacheBackendInterface $cache_backend, ModuleHandlerInterface $module_handler) {
    parent::__construct(
      'Plugin/UiPatterns/PropType',
      $namespaces,
      $module_handler,
      'Drupal\ui_patterns\PropTypeInterface',
      'Drupal\ui_patterns\Annotation\PropType'
    );
    $this->alterInfo('prop_type_info');
    $this->setCacheBackend($cache_backend, 'ui_patterns_prop_type_plugins');
  }

  /**
   * {@inheritdoc}
   */
  public function guessFromSchema(array $prop_schema): ?PropTypeInterface {
    $definition = $this->getDefinitionFromSchema($prop_schema);
    if ($definition !== NULL) {
      return $this->createInstance($definition['id'], []);
    }
    return $this->createInstance('unknown', []);
  }

  /**
   * Returns plugin definitions sorted by priority.
   */
  protected function getSortedDefinitions(): array {
    $definitions = $this->getDefinitions();
    usort($definitions, function ($a, $b) {
      return ($b['priority'] ?? 1) - ($a['priority'] ?? 1);
    });
    return $definitions;
  }

  /**
   * Returns a prop type definition from a JSON schema.
   */
  protected function getDefinitionFromSchema(array $prop_schema): ?array {
    if (isset($prop_schema['$ref']) && str_starts_with($prop_schema['$ref'], "ui-patterns://")) {
      $prop_type_id = str_replace("ui-patterns://", "", $prop_schema['$ref']);
      return $this->getDefinition($prop_type_id);
    }
    $definitions = $this->getSortedDefinitions();
    $compatibilityChecker = \Drupal::service('ui_patterns.schema_compatibility_checker');
    foreach ($definitions as $definition) {
      if ($compatibilityChecker->isCompatible($prop_schema, $definition['schema'])) {
        return $definition;
      }
    }
    return NULL;
  }

  /**
   * {@inheritdoc}
   */
  public function getFallbackPluginId($plugin_id, array $configuration = []) {
    return 'unknown';
  }

}
