<?php

declare(strict_types=1);

namespace Drupal\ui_patterns;

use Drupal\Component\Plugin\PluginBase;
use Drupal\Core\DependencyInjection\DependencySerializationTrait;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Plugin\ContainerFactoryPluginInterface;
use Drupal\Core\Plugin\Context\ContextRepositoryInterface;
use Drupal\Core\Plugin\ContextAwarePluginAssignmentTrait;
use Drupal\Core\Plugin\ContextAwarePluginInterface;
use Drupal\Core\Plugin\ContextAwarePluginTrait;
use Drupal\Core\StringTranslation\StringTranslationTrait;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Base class for source plugins.
 */
abstract class SourcePluginBase extends PluginBase implements
  SourceInterface,
  ContextAwarePluginInterface,
  ContainerFactoryPluginInterface {

  use ContextAwarePluginAssignmentTrait;
  use ContextAwarePluginTrait;
  use StringTranslationTrait;
  use DependencySerializationTrait;

  /**
   * Definition of the targeted prop.
   */
  protected array $propDefinition;

  /**
   * ID of the targeted prop.
   */
  protected string $propId;

  /**
   * The provided plugin contexts.
   *
   * @var array
   */
  protected $context = [];

  /**
   * All gathered plugin contexts.
   *
   * @var array
   */
  protected $gatheredContexts = [];

  /**
   * The plugin settings.
   *
   * @var array
   */
  protected $settings = [];

  /**
   * Whether default settings have been merged into the current $settings.
   *
   * @var bool
   */
  protected $defaultSettingsMerged = FALSE;

  /**
   * {@inheritdoc}
   */
  public function defaultSettings(): array {
    return ['value' => $this->propDefinition['default'] ?? NULL];
  }

  /**
   * {@inheritdoc}
   */
  public static function create(
    ContainerInterface $container,
    array $configuration,
    $plugin_id,
    $plugin_definition
  ) {
    return new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $container->get('context.repository')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function __construct(
    array $configuration,
    $plugin_id,
    $plugin_definition,
    protected ContextRepositoryInterface $contextRepository
  ) {
    parent::__construct($configuration, $plugin_id, $plugin_definition);
    $this->setConfiguration($configuration);
    $this->setDefinedContextValues();
  }

  /**
   * {@inheritdoc}
   */
  public function settingsForm(array $form, FormStateInterface $form_state): array {
    return [];
  }

  /**
   * {@inheritdoc}
   */
  public function getSettings(): array {
    // Merge defaults before returning the array.
    if (!$this->defaultSettingsMerged) {
      $this->mergeDefaults();
    }
    return $this->settings;
  }

  /**
   * {@inheritdoc}
   */
  public function getSetting(string $key): mixed {
    // Merge defaults if we have no value for the key.
    if (!$this->defaultSettingsMerged && !array_key_exists($key, $this->settings)) {
      $this->mergeDefaults();
    }
    return $this->settings[$key] ?? NULL;
  }

  /**
   * Merges default settings values into $settings.
   */
  protected function mergeDefaults() {
    $this->settings += $this->defaultSettings();
    $this->defaultSettingsMerged = TRUE;
  }

  /**
   * {@inheritdoc}
   */
  public function setSettings(array $settings): PluginSettingsInterface {
    $this->settings = $settings;
    $this->defaultSettingsMerged = FALSE;
    return $this;
  }

  /**
   * {@inheritdoc}
   */
  public function setSetting(string $key, mixed $value): PluginSettingsInterface {
    $this->settings[$key] = $value;
    return $this;
  }

  /**
   * {@inheritdoc}
   */
  public function label(): string {
    // Cast the label to a string since it is a TranslatableMarkup object.
    return (string) $this->pluginDefinition['label'];
  }

  /**
   * {@inheritdoc}
   */
  public function getPropValue(): mixed {
    return $this->getSetting('value');
  }

  /**
   * {@inheritdoc}
   */
  public function getConfiguration() {
    return $this->configuration;
  }

  /**
   * {@inheritdoc}
   */
  public function setConfiguration(array $configuration) {
    if (isset($configuration['prop_definition'])) {
      $this->propDefinition = $configuration['prop_definition'];
    }
    if (isset($configuration['prop_id'])) {
      $this->propId = $configuration['prop_id'];
    }
    if (isset($configuration['settings']['context_mapping'])) {
      $configuration['context_mapping'] = $configuration['settings']['context_mapping'];
      unset($configuration['settings']['context_mapping']);
    }
    if (isset($configuration['settings'])) {
      $this->setSettings($configuration['settings']);
    }
    if (isset($configuration['contexts'])) {
      $this->context = $configuration['contexts'];
    }

    $this->configuration = $configuration;
  }

  /**
   * {@inheritdoc}
   */
  public function defaultConfiguration(): array {
    return [];
  }

  /**
   * {@inheritdoc}
   */
  public function getPropId(): string {
    return $this->propId;
  }

  /**
   * {@inheritdoc}
   */
  public function getPropDefinition(): array {
    return $this->propDefinition;
  }

  /**
   * Set values for the defined contexts of this plugin.
   */
  private function setDefinedContextValues() {
    $configuration = $this->getConfiguration();
    // Fetch the available contexts.
    $available_contexts = $this->contextRepository->getAvailableContexts();

    $available_runtime_contexts = $this->context;
    // Ensure that the contexts have data by getting corresponding runtime
    // contexts.
    $available_runtime_contexts += $this->contextRepository->getRuntimeContexts(
      array_keys($available_contexts)
    );
    $plugin_context_definitions = $this->getContextDefinitions();
    $this->gatheredContexts = $available_runtime_contexts;
    foreach ($plugin_context_definitions as $name => $plugin_context_definition) {
      // Identify and fetch the matching runtime context, with the plugin's
      // context definition.
      $matches = $this->contextHandler()
        ->getMatchingContexts(
          $available_runtime_contexts,
          $plugin_context_definition
        );
      $context_mapping = $configuration['context_mapping'] ?? [];
      if (isset($context_mapping[$name]) && isset($matches[$context_mapping[$name]])) {
        $matching_context = $matches[$context_mapping[$name]];
      }
      else {
        $matching_context = reset($matches);
      }
      if ($matching_context) {
        $this->setContextValue($name, $matching_context->getContextValue());
      }
    }
  }

  /**
   * {@inheritdoc}
   */
  public function alterComponent(array $element): array {
    return $element;
  }

}
