<?php

declare(strict_types=1);

namespace Drupal\ui_patterns;

use Drupal\Core\Cache\CacheBackendInterface;
use Drupal\Core\Extension\ModuleHandlerInterface;
use Drupal\Core\Plugin\Context\ContextAwarePluginManagerInterface;
use Drupal\Core\Plugin\Context\ContextAwarePluginManagerTrait;
use Drupal\Core\Plugin\DefaultPluginManager;
use Drupal\ui_patterns\Annotation\Source;

/**
 * Source plugin manager.
 */
class SourcePluginManager extends DefaultPluginManager implements ContextAwarePluginManagerInterface {

  use ContextAwarePluginManagerTrait;

  /**
   * Constructs the object.
   */
  public function __construct(
    \Traversable $namespaces,
    CacheBackendInterface $cache_backend,
    ModuleHandlerInterface $module_handler,
    protected PropTypePluginManager $prop_type_manager,
  ) {
    parent::__construct(
      'Plugin/UiPatterns/Source',
      $namespaces,
      $module_handler,
      SourceInterface::class,
      Source::class
    );
    $this->alterInfo('ui_patterns_source_info');
    $this->setCacheBackend($cache_backend, 'ui_patterns_source_plugins');
  }

  /**
   * Returns source definitions for a prop type.
   */
  public function getDefinitionsForPropType(string $prop_type_id, array $contexts = []): array {
    $definitions = $this->getDefinitionsForContexts($contexts);
    $definitions = array_filter($definitions, function ($definition) use ($prop_type_id) {
      return in_array($prop_type_id, $definition['prop_types']);
    });
    return $definitions;
  }

  /**
   * Returns the default source identifier for a prop type.
   */
  public function getPropTypeDefault(string $prop_type_id, array $contexts = []): ?string {
    // First try with prop type default source.
    $prop_type_definition = $this->prop_type_manager->getDefinition($prop_type_id);
    $source_id = $prop_type_definition["default_source"] ?? NULL;
    if ($source_id && $this->isApplicable($source_id, $contexts)) {
      return $source_id;
    }
    // If no default source, take the first one.
    $definitions = $this->getDefinitionsForPropType($prop_type_id, $contexts);
    $source_id = array_key_first($definitions) ?? NULL;
    if ($source_id && $this->isApplicable($source_id, $contexts)) {
      return $source_id;
    }
    return NULL;
  }

  /**
   * Build plugin configuration.
   */
  public static function buildPluginConfiguration(string $prop_id, array $prop_definition, array $settings = [], array $source_contexts = []): array {
    $context_mapping = $settings['source']['context_mapping'] ?? [];
    unset($settings['source']['context_mapping']);
    return [
      'prop_id' => $prop_id,
      'prop_definition' => $prop_definition,
      'contexts' => $source_contexts,
      'settings' => $settings['source'] ?? [],
      'context_mapping' => $context_mapping,
    ];
  }

  /**
   * Creates a plugin instances with the same configuration.
   *
   * @param array $plugin_ids
   *   The source plugin identifiers.
   * @param array $configuration
   *   An array of configuration.
   *
   * @return array
   *   A list of fully configured plugin instances.
   */
  public function createInstances(array $plugin_ids, array $configuration): array {
    return array_map(
      function ($plugin_id) use ($configuration) {
        return $this->createInstance($plugin_id, $configuration);
      },
      $plugin_ids,
    );
  }

  /**
   * Check if the source is matching the specified context.
   *
   * @param string $source_id
   *   The source plugin identifier.
   * @param array $contexts
   *   An array of contexts.
   *
   * @return bool
   *   Is the source applicable.
   */
  public function isApplicable(string $source_id, array $contexts): bool {
    // @todo use a method of the plugin instead?
    $definitions = $this->getDefinitionsForContexts($contexts);
    return isset($definitions[$source_id]);
  }

}
